# ESA (C) 2000-2022
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# XMMextractor_verify


import numpy as np
import os
import sys
import glob
from logger import get_logger
from pysas.pyutils import pyutils



############################################
############## EPIC images ################
###########################################
def test_existance_epic_clean_image(obs, expo,type=None):
    '''
    Checks and retunr EPIC gti filetered full image
    Args:
       obs: observation objects
       exposure: exposure identifier
    '''
    
    EPNlog = get_logger(os.path.join(obs.working_dir, 'EPNlog'))
    wd = os.getcwd()
    try:
        os.chdir(obs.images_dir)
    except FileNotFoundError:
        return (0, 'No images directory')
    
    if type == 'full':
        inst = None
        if  expo[0:2].lower() == 'pn':
            inst = 'EPN'
        elif expo[0:2].lower() == 'm1':
            inst = 'EMOS1'
        elif expo[0:2].lower() == 'm2':
            inst = 'EMOS2'

        image_file =  glob.glob('*'+inst + '_' + expo[2:6] + '_Imaging_Image.ds')
    else:
        image_file =  glob.glob( expo[0:2].lower() + '_' + expo[2:6] + '_IMAGING_image_full.fits')

    os.chdir(wd)
    if len(image_file) != 1:
        EPNlog.info('#>  Image file not found')
        return(0, 'No file')
    else:
        return(True, obs.images_dir + '/' + image_file[0])

############################################
################ EPN files #################
############################################

def test_existance_pn_eventfiles(obs, exposure =None,type='EVT'):
    '''
    Checks if for a given expid exist event lists.
    
    Args:
        obs: the observation object.
        exposure: an exposure.
    
    Output:
        tuple. First element one (file) or zero (no file). Second element,
    a string with the file or 'No file'.
    '''

    EPNlog = get_logger(os.path.join(obs.working_dir, 'EPNlog'))

    wd = os.getcwd()

    expid = ''
    if exposure is not None:
        if isinstance(exposure, str):
            expid = exposure[-4:]
        else:
            expid = exposure.expid[-4:]

    
    EPIC_evlist = ""
    if type == 'EVT':
        if not os.getcwd().endswith('EPN'):
            try:
                os.chdir(obs.EPN_dir)
            except FileNotFoundError:
                return (0, 'No EPN directory')

        EPIC_evlist = glob.glob('*EPN*' + expid + '*ImagingEvts.ds')

    elif type == 'CLEAN':
        if not os.getcwd().endswith('GTI'):
            try:
                os.chdir(obs.GTI_dir)
            except FileNotFoundError:
                return (0, 'No EPN directory')
        EPIC_evlist = glob.glob('*pn*' + expid + '*ImagingEvts_events_gtifiltered.fits')


    if len(EPIC_evlist) > 0:
        index = 0
        duration = []
        EPNlog.info('PN Image Event List found: {}'.format(EPIC_evlist))
        for i in EPIC_evlist:
            obj_duration = pyutils.get_key_word(i, 'DURATION', 1)
            duration.append(obj_duration)

        max_index = np.argmax(duration)
        pn_list = EPIC_evlist[max_index]
        EPNlog.info('Using event file with longest Observation Duration: {}'.format(pn_list))
        os.chdir(wd)
        if type == 'EVT':
            return(True, obs.EPN_dir + '/' + pn_list)
        else:
            return(True, obs.GTI_dir + '/' + pn_list)
    else:
        EPNlog.info('#> PN Imaging Event List not found')
        os.chdir(wd)
        return(0, 'No file')


def test_existance_pn_timing_eventfiles(obs,exposure = None,type='EVT'):
    '''
    Checks if for a given expid exist timing event lists.
    
    Args:
        obs: the observation object.
        exposure: an exposure.
    
    Output: 
        tuple. First element one (file) or zero (no file). Second element,
    a string with the file or 'No file'.
    '''

    EPNlog = get_logger(os.path.join(obs.working_dir, 'EPNlog'))

    wd = os.getcwd()

    if exposure is not None:
        if isinstance(exposure, str):
            expid = exposure[-4:]
        else:
            expid = exposure.expid[-4:]

    
    EPIC_evlist = ""
    if type == 'EVT':
        if not os.getcwd().endswith('EPN'):
            try:
                os.chdir(obs.EPN_dir)
            except FileNotFoundError:
                return (0, 'No EPN directory')
            
        EPIC_evlist = glob.glob('*EPN*' + expid + '*TimingEvts.ds')

    elif type == 'CLEAN':
        if not os.getcwd().endswith('GTI'):
            try:
                os.chdir(obs.GTI_dir)
            except FileNotFoundError:
                return (0, 'No directory')
        EPIC_evlist = glob.glob(exposure+'*_events_gti*Filtered.fits')


    if len(EPIC_evlist) > 0:
        EPNlog.info('PN Timing Event List found: {}'.format(EPIC_evlist))
        
        print(EPIC_evlist[0])
        
        EPNlog.info('Using event file with longest Observation Duration: {}'.format(EPIC_evlist[0]))
        os.chdir(wd)
        if type == 'EVT':
            return(True, obs.EPN_dir + '/' + EPIC_evlist[0])
        else:
            return(True, obs.GTI_dir + '/' + EPIC_evlist[0])

    else:
        EPNlog.info('#> PN Timing List not found')
        os.chdir(wd)
        return(0, 'No file')


    
#############################################
################# MOS files #################
#############################################

def test_existance_EMOS(obs, exposure,type='EVT'):
    ''' 
    Checks if for a given expid exist event lists (EMOS1).
    
    Args:
        obs: an observation object.
        exposure: an exposure.
    
    Output: 
        tuple. First element one (file) or zero (no file). Second element,
    a string with the file or 'No file'.
    '''

    EMOSlog = get_logger(os.path.join(obs.working_dir, 'EMOSlog'))

    wd = os.getcwd()
    instrument = ""
    if(exposure[0:2] == 'M1'):
        instrument = 'EMOS1'
    elif(exposure[0:2] == 'M2'):
        instrument = 'EMOS2'
    else:
        EMOSlog.error('Instrument not found')
        
    exp_id = exposure[2:6]
    EMOS_evlist=""
    dir=''
    if type == 'EVT':
        if not os.getcwd().endswith('EMOS'):
            try:
                os.chdir(obs.MOS_dir)
            except FileNotFoundError:
                return (0, 'No directory')
        EMOS_evlist = glob.glob('*'+instrument+'*' + exp_id + '*ImagingEvts.ds')
        dir = obs.MOS_dir
        
    if type == 'CLEAN':
        if not os.getcwd().endswith('GTI'):
            try:
                os.chdir(obs.GTI_dir)
            except FileNotFoundError:
                return (0, 'No directory')

        EMOS_evlist = glob.glob('*'+exposure[0:2].lower() + '*'+ exp_id + '*_events_gti*Filtered.fits')
        dir = obs.GTI_dir        
        

        
    if len(EMOS_evlist) > 0:
        index = 0
        duration = []
        EMOSlog.info('#> {} Image Event List found: {}'.format(instrument,EMOS_evlist))
        for i in EMOS_evlist:
            obj_duration = pyutils.get_key_word(i, 'DURATION', 1)
            duration.append(obj_duration)

        max_index = np.argmax(duration)
        EMOS_list = EMOS_evlist[max_index]

        EMOSlog.info('      Using event file with longest Observation Duration: {}'.format(EMOS_list))
        os.chdir(wd)
        return(1, dir + '/' + EMOS_list)
    else:
        EMOSlog.info('#> EMOS1 Imaging Event List not found')
        os.chdir(wd)
        return(0, 'No file')


def test_existance_EMOS_timing(obs, exposure,type='EVT'):
    ''' 
    Checks if for a given expid exist timing event lists (EMOS)
    
    Args:
        obs: an observation object.
        exposure: an exposure.
    
    Output: 
        tuple. First element one (file) or zero (no file). Second element,
    a string with the file or 'No file'.
    '''
    
    EMOSlog = get_logger(os.path.join(obs.working_dir, 'EMOSlog'))
    
    wd = os.getcwd()

    instrument = ""
    if(exposure[0:2] == 'M1'):
        instrument = 'EMOS1'
    elif(exposure[0:2] == 'M2'):
        instrument = 'EMOS2'
    else:
        EMOSlog.error('Instrument not found')
        
    exp_id = exposure[2:6]
    EMOS_timeevlist=""
    dir=''
    if type == 'EVT':
        if not os.getcwd().endswith('EMOS'):
            try:
                os.chdir(obs.MOS_dir)
            except FileNotFoundError:
                return (0, 'No directory')
        EMOS_timeevlist = glob.glob('*'+instrument+'*' + exp_id + '*TimingEvts.ds')
        dir = obs.MOS_dir

    if type == 'CLEAN':   
        if not os.getcwd().endswith('GTI'):
            try:
                os.chdir(obs.GTI_dir)
            except FileNotFoundError:
                return (0, 'No directory')
        EMOS_timeevlist = glob.glob(exposure+ '*_events_gti*Filtered.fits')
        dir = obs.GTI_dir

    if len(EMOS_timeevlist) > 0:
        index = 0
        duration = []
        EMOSlog.info('#> {} timing Event List found: {}', instrument, EMOS_timeevlist)
        for i in EMOS_timeevlist:
            obj_duration = pyutils.get_key_word(i, 'DURATION', 1)
            duration.append(obj_duration)

        max_index = np.argmax(duration)

        EMOS_timelist = EMOS_timeevlist[max_index]

        EMOSlog.info('      Using event file with longest Observation Duration: {}', EMOS_timelist)
        os.chdir(wd)
        return(1, dir + '/' + EMOS_timelist)
    else:
        EMOSlog.info('#> EMOS1 Timing Event List not found')
        os.chdir(wd)
        return(0, 'No file')


#############################################
################# RGS files #################
#############################################

def test_existance_RGS(obs,expo):
    '''
    Checks if exist event lists (RGS1).

    Args:
        obs: the observation object.
        expo: RGS instrument and exposures

    Output:
        tuple. First element one (file) or zero (no file). Second element,
    a string with the file or 'No file'.
    '''

    RGSlog = get_logger(os.path.join(obs.working_dir, 'RGSlog'))

        
    if not os.getcwd().endswith('RGS'):
        try:
            os.chdir(obs.RGS_dir)
        except FileNotFoundError:
            return (0, 'No directory')

    RGS_evlist = glob.glob('*'+expo+'*EVENLI0000.FIT')
    if len(RGS_evlist) > 0:
        RGSlog.debug('   #> RGS1 Event List found')
        RGSlog.info(RGS_evlist)
        return (1, os.getcwd() + '/' + RGS_evlist[0])
    else:
        RGSlog.info('#> RGS1 Event List not found')
        return(0, 'No file')


def test_existance_SRC_RGS(obs,expo):
    '''
    Checks if exist source files (RGS1).

    Args:
        obs: an observation object.
        expo: instrument plus exposure id
    
    Output:
        tuple. First element one (file) or zero (no file). Second element,
    a string with the file or 'No file'.
    '''

    RGSlog = get_logger(os.path.join(obs.working_dir, 'RGSlog'))

    if not os.getcwd().endswith('RGS'):
        try:
            os.chdir(obs.RGS_dir)
        except FileNotFoundError:
            return (0, 'No directory')

    RGS_srclist = glob.glob('*'+expo+'*SRCLI_0000.FIT')
    if len(RGS_srclist) > 0:
        RGSlog.info('   #> RGS1 Source List found')
        RGSlog.info(RGS_srclist)
        return (1, os.getcwd() + '/' + RGS_srclist[0])
    else:
        RGSlog.info('#> RGS Source List not found')
        return(0, 'No file')




#############################################
################# OM  files #################
#############################################

def test_OM_event_files(obs):
    '''
    Checks for event list produced by omichain.

    Output:
        a tuple, with 0 or 1 (not found, found) and 'No file' or the name
    of the file.
    '''

    omlog = get_logger(os.path.join(obs.working_dir, 'OMlog'))

    if not os.getcwd().endswith('OM'):
        try:
            os.chdir(obs.OM_dir)
        except FileNotFoundError:
            return (0, 'No OM directory')

    OM_event_list = glob.glob('P*OM*SIMAGE*.FIT')
    if len(OM_event_list) > 0:
        omlog.info('#> OM Event List found [omichain]')
        omlog.info(OM_event_list)
        return (1, os.path.join(obs.OM_dir, OM_event_list[0]))
        #if len(OM_event_list) > 1:
        #    omlog.info('Using only {}'.format(OM_event_list[0]))
        #    OM_event = OM_event_list[0]
        #    return (1, os.getcwd() + '/' + OM_event)
    else:
        omlog.info('#> OM Event List not found for omichain processing.')
        return (0, 'No file')


def test_OMF_event_files(obs):
    '''
    Checks for event list produced by omfchain.

    Output:
        a tuple, with 0 or 1 (not found, found) and 'No file' or the name
    of the file.
    '''

    omlog = get_logger(os.path.join(obs.working_dir, 'OMlog'))

    if not os.getcwd().endswith('OM'):
        try:
            os.chdir(obs.OM_dir)
        except FileNotFoundError:
            return (0, 'No directory')

    OM_event_list = glob.glob('*OM*EVLIST*.FIT')
    if len(OM_event_list) > 0:
        omlog.info('#> OM Event List found [omfchain]')
        omlog.info(OM_event_list)
        return (1, os.path.join(obs.OM_dir, OM_event_list[0]))
        #if len(OM_event_list) > 1:
        #    mlog.log('info', 'Using only {}'.format(OM_event_list[0]))
        #    OM_event = OM_event_list[0]
        #    return (1, os.getcwd() + '/' + OM_event)
    else:
        omlog.info('#> OM Event List not found for omfchain processing.')
        return (0, 'No file')


def test_OMG_event_files(obs):
    '''
    Checks for event list produced by omgchain.

    Output:
        a tuple, with 0 or 1 (not found, found) and 'No file' or the name
    of the file.
    '''

    omlog = get_logger(os.path.join(obs.working_dir, 'OMlog'))

    if not os.getcwd().endswith('OM'):
        try:
            os.chdir(obs.OM_dir)
        except FileNotFoundError:
            return (0, 'No directory')

    OM_event_list = glob.glob('p*OM*RIMAGE*.FIT')
    if len(OM_event_list) > 0:
        omlog.info('#> OM Event List found [omgchain]')
        omlog.info(OM_event_list)
        return (1, os.path.join(obs.OM_dir, OM_event_list[0]))
        #if len(OM_event_list) > 1:
        #    xmmlog.info('    Using only {}'.format(OM_event_list[0]))
        #    OM_event = OM_event_list[0]
        #    return (1, os.getcwd() + '/' + OM_event)
    else:
        omlog.info('  #> OM Event List not found for omgchain processing.')
        return (0, 'No file')


def test_OMSRC_files(obs):
    '''
    Checks for Source files.

    Output:
        a tuple, with 0 or 1 (not found, found) and 'No file' or the name
    of the file.
    '''

    omlog = get_logger(os.path.join(obs.working_dir, 'OMlog'))

    if not os.getcwd().endswith('OM'):
        try:
            os.chdir(obs.OM_dir)
        except FileNotFoundError:
            return (0, 'No directory')

    OM_event_list = glob.glob('*P*OM*SRLI*.FIT')
    if len(OM_event_list) > 0:
        omlog.info('#> OM Source List found')
        omlog.info(OM_event_list)
        return (1, os.path.join(obs.OM_dir, OM_event_list[0]))
        #if len(OM_event_list) > 1:
        #    omlog.log('    Using only {}'.format(OM_event_list[0]))
        #    OM_event = OM_event_list[0]
        #return (1, os.getcwd() + '/' + OM_event)
    else:
        omlog.info('   #> OM Source List not found.')
        return (0, 'No file')


def test_OMSRCCombo_files(obs):
    '''
    Checks for Source combo files.

    Output:
        a tuple, with 0 or 1 (not found, found) and 'No file' or the name
    of the file.
    '''
    omlog = get_logger(os.path.join(obs.working_dir, 'OMlog'))

    if not os.getcwd().endswith('OM'):
        try:
            os.chdir(obs.OM_dir)
        except FileNotFoundError:
            return (0, 'No directory')

    OM_event_list = glob.glob('*P*OMCOMBOBSMLI*.FIT')
    if len(OM_event_list) > 0:
        omlog.info('#> OM Source Combo List found')
        omlog.info(OM_event_list)

        if len(OM_event_list) > 1:
            omlog.info('Using only {}'.format(OM_event_list[0]))
            OM_event = OM_event_list[0]
            return (1, os.path.join(obs.OM_dir, OM_event))
            #return (1, os.getcwd() + '/' + OM_event)
        else:
            OM_event = OM_event_list[0]
            omlog.info('Using {}'.format(OM_event))
            return (1, os.path.join(obs.OM_dir, OM_event))
            #return (1, os.getcwd() + '/' + OM_event)
    else:
        omlog.info('#> OM Source Combo List not found.')
        return (0, 'No file')



#############################################
############## Spectra   files ##############
#############################################


def check_spectra_files(dir,expo):
    '''
    Check for the existence of EPIC spectra files.
    
    Args:
        dir: spectra directory
        expo: exposure object

    Output: 
        1 if EPIC file is found, 0 otherwise.
    '''
    
    inst = None
    if  expo[0:2].lower() == 'pn':
        inst = 'EPN'
    elif expo[0:2].lower() == 'm1':
        inst = 'EMOS1'
    elif expo[0:2].lower() == 'm2':
        inst = 'EMOS2'

    srcfile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*source_spectrum.fits')
    bkgfile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*background_spectrum.fits')
    arffile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*.arf')    
    rmffile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*.rmf')
    
    if len(srcfile) != 0 and  len(bkgfile) != 0 and len(arffile) != 0 and len(rmffile) != 0 :
        return (1,[srcfile[0],bkgfile[0],arffile[0],rmffile[0]])
    else:
        return  (0, 'No file')


################################################
############## Lightcurve   files ##############
################################################


def check_lightcurve_files(dir,expo):
    '''
    Check for the existence of EPIC spectra files.
    
    Args:
        dir: lightcurve directory
        expo: exposure object

    Output: 
        1 if EPIC files are found, 0 otherwise.
    '''
    
    inst = None
    if  expo[0:2].lower() == 'pn':
        inst = 'EPN'
    elif expo[0:2].lower() == 'm1':
        inst = 'EMOS1'
    elif expo[0:2].lower() == 'm2':
        inst = 'EMOS2'

    srcFile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*source.lc')
    bkgFile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*bkg.lc')
    bkgSubstractedFile = glob.glob(dir + '/*'+inst+'*'+expo[2:6]+'*sourcebkgsubtracted.lc')    

    
    if len(srcFile) != 0 and  len(bkgFile) != 0 and len(bkgSubstractedFile) != 0 :
        return (1,[srcFile[0],bkgFile[0],bkgSubstractedFile[0]])
    else:
        return  (0, 'No file')


################################################
############## Lightcurve   files ##############
################################################


def check_edetect_images(dir,expo):
    '''
    Check for the existence of EPIC spectra files.
    
    Args:
        dir: lightcurve directory
        expo: exposure object

    Output: 
        1 if EPIC files are found, 0 otherwise.
    '''
    
    fullFile = glob.glob(dir + '/*'+expo[0:2].lower()+'*'+expo[2:6]+'*full.fits')
    enerFiles = glob.glob(dir + '/*'+expo[0:2].lower()+'*'+expo[2:6]+'*image_b?.fits')

    
    if len(fullFile) != 0 and  len(enerFiles) != 0  :
        return (1,[fullFile[0],enerFiles])
    else:
        return  (0, 'No file')


def check_edetect_files(dir,expo):
    '''
    Check for the existence of EPIC spectra files.
    
    Args:
        dir: lightcurve directory
        expo: exposure object

    Output: 
        1 if EPIC files are found, 0 otherwise.
    '''
    

    ebox_l_file = glob.glob(dir + '/*'+expo[0:2].lower()+'*'+expo[2:6]+'*eboxlist_l.fits')
    ebox_m_file = glob.glob(dir + '/*'+expo[0:2].lower()+'*'+expo[2:6]+'*eboxlist_m.fits')
    emllist_file = glob.glob(dir + '/*'+expo[0:2].lower()+'*'+expo[2:6]+'*emllist.fits')

    
    if len(ebox_l_file) != 0 and  len(ebox_m_file) != 0 and len(emllist_file) != 0 :
        return (1,[ebox_l_file[0],ebox_m_file[0],emllist_file[0]])
    else:
        return  (0, 'No file')
    




    
#############################################
################# GTI files #################
#############################################

def check_GTI_files(gti_dir, expo):
    '''
    Check for the existance of EPIC timing event files (non-filtered).
    
    Args:
        gti_dir: GTI directory
        instr: the instrument.
        exposure: the exposure.

    Output:
        1 if found, 0 otherwise.
        filename
    '''
    
    files = glob.glob(gti_dir + '/'+expo+'_*_*_gti*.fits')

    if len(files) != 0:
        return (1,files)
    else:
        return  (0, 'No file')

    
def check_flaring_event_files(dir,expo):
    '''
    Check for the existance of EPIC events files cleaned by flaring bkg (filtered).
    
    Args:
        dir: GTI directory
        expo: exposure identifier.

    Output:
        1 if found, 0 otherwise.
        filename
    '''
    files=  glob.glob(dir+'/'+ expo[0:2].lower() + '_' + expo[2:6] + '_ImagingEvts_events_gtiFlaringFiltered.fits')


    if len(files) != 0:
        return (1,files)
    else:
        return  (0, 'No file')
    




#!/usr/bin/env python

#	ESA (C) 2000-2018 
#	This file is part of ESA's XMM-Newton Scientific Analysis System
#	(SAS).
#
#	SAS is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	SAS is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	   
#	You should have received a copy of the GNU General Public License
#	along with SAS.  If not, see <http://www.gnu.org/licenses/>.


from PyQt5.QtWidgets import QFrame
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QVBoxLayout, QWidget,QLineEdit
from PyQt5.QtCore import QRect,Qt,pyqtSlot
from functools import partial
from PyQt5.Qt import QScrollArea, QGroupBox
import logging

class controlPanel:
    #on_click = pyqtSignal()
    
    def __init__(self,procEPNDict,procEMOS1Dict,procEMOS2Dict,procRGS1Dict,\
                 procRGS2Dict,procOMDict):

        self.logger = logging.getLogger('xmmextractorGUI')
        self.logger.info("Creating Control Panel")

        self.procEPNDict = procEPNDict
        self.procEMOS1Dict = procEMOS1Dict
        self.procEMOS2Dict = procEMOS2Dict
        self.procRGS1Dict = procRGS1Dict
        self.procRGS2Dict = procRGS2Dict
        self.procOMDict = procOMDict
        #self.tabObservationWidget = tabObservationWidget
        self.EPNExpButtonDict= {}
        self.height = 0
        self.btnWidth = 110
        self.btnHeight = 24
    
    def createPanel(self): 
        self.logger.info("Creating Control Panel")
    
    def getFrame(self):        
               
        self.EPNControlFrame = QtWidgets.QFrame()
        self.EPNFrameLayout = QVBoxLayout()
        self.EPNControlFrame.setMinimumSize(100,1000)
        self.EPNControlFrame.setLayout(self.EPNFrameLayout)
        self.EPNControlFrame.setGeometry(QRect(0,25,650,300))
        
        
        self.createEPNControlPanel(self.procEPNDict)
        
        self.createEPNControlPanel(self.procEMOS1Dict)
        self.createEPNControlPanel(self.procEMOS2Dict)
        self.createEPNControlPanel(self.procRGS1Dict)
        self.createEPNControlPanel(self.procRGS2Dict)
        self.createEPNControlPanel(self.procOMDict)
                
        self.scroll = QScrollArea()        
        self.scroll.setWidgetResizable(True) 
        self.scroll.setVisible(True)
        self.scroll.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        
        self.scroll.setWidget(self.EPNControlFrame)
        #self.scroll.setWidget(self.totalFrame)        
        return self.scroll
        #return self.controlFrame
    
    

    def createEPNControlPanel(self,instDict):
        
        if len(instDict) == 2:
            return

        j = (((len(instDict ) - 2)*20)/2. - 10) + self.height               
        self.EPNButton = QtWidgets.QPushButton(self.EPNControlFrame)
        self.EPNButton.setGeometry(QRect(0,round(j),self.btnWidth,self.btnHeight))
        inst = instDict['Instrument']
        instButtonName = inst+'Button'
        self.EPNButton.setObjectName(instButtonName)
        self.EPNButton.setText(inst)
        self.EPNButton.setCheckable(True);        
        if instDict['Processing'] == "yes":
            self.EPNButton.setChecked(True)
            instrumentFlag = True
        else:
            self.EPNButton.setChecked(False)
            instrumentFlag = False            
        self.EPNButton.toggle()
        self.EPNButton.clicked.connect(partial(self.EPN_on_click,inst,instDict))
        self.EPNExpButtonDict[inst] = self.EPNButton
        #EPNButtongList = []
       
        
        j = self.height
        for key,val in instDict.items():
            RGSFlag = True
            OMFlag = True
            i = self.btnWidth
            EPNExpButtonDict = {}                            
            if isinstance(val,dict):
                self.ExpButton = QtWidgets.QPushButton(key, self.EPNControlFrame)
                self.ExpButton.setGeometry(QRect(i,j,self.btnWidth,self.btnHeight))                
                self.ExpButton.setObjectName(key)
                self.ExpButton.setText(key)                
                self.ExpButton.setCheckable(True);
                
                if instDict[key]['Process'] == "yes" and instrumentFlag == True:
                    self.ExpButton.setChecked(True)
                else:
                    self.ExpButton.setChecked(False)
                self.ExpButton.toggle()
                self.ExpButton.clicked.connect(partial(self.Exp_on_click,key,instDict))
                self.EPNExpButtonDict[key] = self.ExpButton
                #print ('EXPOSURE ', key)
                for expKey,expInfo in val.items():
                    if expKey != "Process":
                        i = i + self.btnWidth                        

                    if expKey == "pileup":
                        i = i - self.btnWidth

                    #For RGS we do not have "yet" the possibility to create GTI
                    if (inst == "RGS1" or inst == "RGS2")  and  (expKey != "Process" and\
                      expKey != "EventList"  and RGSFlag == True):                        
                        for x in range(0,2):                            
                            self.ExpInfoButton = QtWidgets.QPushButton(self.EPNControlFrame)
                            self.ExpInfoButton.setGeometry(QRect(i,j,self.btnWidth,self.btnHeight))
                            self.ExpInfoButton.setObjectName(expInfo)
                            self.ExpInfoButton.setText("N/A")                            
                            self.ExpInfoButton.setEnabled(False);
                            i = i + self.btnWidth
                            RGSFlag = False

                    if inst == "OM"  and  (expKey != "Process" and\
                    expKey != "EventList" and OMFlag == True):
                        
                        for x in range(0,4):                            
                            self.ExpInfoButton = QtWidgets.QPushButton(self.EPNControlFrame)
                            self.ExpInfoButton.setGeometry(QRect(i,j,self.btnWidth,self.btnHeight))
                            self.ExpInfoButton.setObjectName(expInfo)
                            self.ExpInfoButton.setText("N/A")                            
                            self.ExpInfoButton.setEnabled(False);
                            i = i + self.btnWidth
                            OMFlag = False
                            

                    
                    #print(expKey, ' ',expInfo)
                    self.ExpInfoButton = QtWidgets.QPushButton(self.EPNControlFrame)
                    self.ExpInfoButton.setGeometry(QRect(i,j,self.btnWidth,self.btnHeight))
                    self.ExpInfoButton.setObjectName(expInfo)
                    self.ExpInfoButton.setText(expInfo)
                    self.ExpInfoButton.setCheckable(True);


                    if instDict[key][expKey] == "yes" and instrumentFlag == True:
                        self.ExpInfoButton.setChecked(True)                        
                    else:
                        self.ExpInfoButton.setChecked(False)
                        self.ExpInfoButton.setText("no")
                        instDict[key][expKey] = "no"


                    self.ExpInfoButton.toggle()
                    self.ExpInfoButton.clicked.connect(partial(self.ExpInfo_on_click,key,expKey,instDict))
                    expVal = key+"_"+expKey
                    self.EPNExpButtonDict[expVal] = self.ExpInfoButton
                    if expKey == "omchain":
                        self.ExpInfoButton.setText("N/A")
                        self.ExpInfoButton.setEnabled(False);
                        
                    if expKey == "Process" or expKey == "pileup" :
                        self.ExpInfoButton.hide()
                    
                    
                #EPNButtongList.append(EPNExpButtonDict)
                j=j+self.btnHeight
            #else:               
                #print (key,' ',val)
        
        self.height = j
               
    
    def getEPNDict(self):
        return self.procEPNDict
    
    def setFrame(self,newFrame):
        self.logger.info("UPDATING CONTROL FRAME")
        self.controlFrame = newFrame
        self.controlFrame.update()
        
    def getFrame2(self):
        return self.controlFrame
        
    def updateControlPanel(self,procEPNDict,procEMOS1Dict,procEMOS2Dict,
                           procRGS1Dict,procRGS2Dict,procOMDict):
        self.logger.info("Updating Buttons...")
        self.procEPNDict = procEPNDict
        self.procEMOS1Dict = procEMOS1Dict
        self.procEMOS2Dict = procEMOS2Dict
        self.procRGS1Dict = procRGS1Dict
        self.procRGS2Dict = procRGS2Dict
        self.procOMDict = procOMDict
        '''
        for instrument in Instruments:
            print ("Instrument "+str(instrument.name))
            #if instrument == Instruments.EPN:
            if self.procEPNDict['Processing'] == 'yes':
                self.EPNExpButtonDict[instrument.name].setChecked(False)
        self.changeButtonState(self.procEPNDict,False)
        self.changeButtonState(self.procEMOS1Dict,False)
        self.changeButtonState(self.procEMOS2Dict,False)
        self.changeButtonState(self.procRGS1Dict,False)
        self.changeButtonState(self.procRGS2Dict,False)
        #self.changeButtonState(self.procOMDict,False)
        
        if self.procEPNDict['Processing'] == 'yes':
            self.EPNExpButtonDict['EPN'].setChecked(False)        
        self.changeButtonState(self.procEPNDict,False)
        '''
        #print(self.procEPNDict)
    # Create the actions
    @pyqtSlot()
    def EPN_on_click(self,inst,instrumentDict):
        self.logger.info("EPN button clicked")
        self.logger.info("Status: "+str(self.EPNExpButtonDict[inst].isChecked()))    
        #Get all exposures               
        if self.EPNExpButtonDict[inst].isChecked() == True :
            instrumentDict["Processing"] = "no"
            self.changeButtonState(instrumentDict,True)
        else:
            instrumentDict["Processing"] = "yes"
            self.changeButtonState(instrumentDict,False)
    
    def Exp_on_click(self,a,instrumentDict):        
        self.logger.info("Exp on-click "+a)
        self.logger.info("Status "+str(self.EPNExpButtonDict[a].isChecked()))
        if self.EPNExpButtonDict[a].isChecked() == True :
            instrumentDict[a]["Process"] = "no"
            self.changeExpButtonState(instrumentDict,a,True)
        else:
            instrumentDict[a]["Process"] = "yes"
            self.changeExpButtonState(instrumentDict,a,False)
        
    def ExpInfo_on_click(self,key,val,instrumentDict):
        self.logger.info("Exp Info on-click "+key+" "+val)

        expVal=key+"_"+val
        self.logger.info("Status: "+str(self.EPNExpButtonDict[expVal].isChecked()))
        if self.EPNExpButtonDict[expVal].isChecked() == True:
            instrumentDict[key][val] = "no"
            self.EPNExpButtonDict[expVal].setText("no")
        else:            
            instrumentDict[key][val] = "yes"
            self.EPNExpButtonDict[expVal].setText("yes")
            instrumentDict[key]['Process'] = "yes"
            instrumentDict['Processing'] = "yes"
            self.EPNExpButtonDict[key].setChecked(False)
            self.EPNExpButtonDict[instrumentDict['Instrument']].setChecked(False)
    
    def changeButtonState(self,instrumentDict,valState):
        self.logger.info("VAL STATE "+ str(valState))
        for key,val in instrumentDict.items(): 
            if isinstance(val,dict):
                if valState == True:
                    instrumentDict[key]["Process"] = "no"                    
                else:
                    instrumentDict[key]["Process"] = "yes"
                    
                self.EPNExpButtonDict[key].setChecked(valState)
                for expKey,expInfo in val.items():
                    expVal = key+"_"+expKey  
                    if valState == True:
                        instrumentDict[key][expKey] = "no"
                        self.EPNExpButtonDict[expVal].setText("no")
                    else:
                        instrumentDict[key][expKey] = "yes"
                        self.EPNExpButtonDict[expVal].setText("yes")                    
                    self.EPNExpButtonDict[expVal].setChecked(valState)
            #else:
                #print(key,' ',val)
    def changeExpButtonState(self,instrumentDict,a,valState):                    
        for key,val in instrumentDict.items(): 
            if isinstance(val,dict):
                for expKey,expInfo in val.items():
                    expVal = a+"_"+expKey
                    if valState == True:
                        instrumentDict[a][expKey] = "no"
                        self.EPNExpButtonDict[expVal].setText("no")                        
                    else:
                        instrumentDict[a][expKey] = "yes"
                        instrumentDict['Processing'] = "yes"
                        self.EPNExpButtonDict[expVal].setText("yes")
                        self.EPNExpButtonDict[instrumentDict['Instrument']].setChecked(valState)
                    
                    self.EPNExpButtonDict[expVal].setChecked(valState)
                
        
    def activateSrcDetection(self,state):
        self.logger.info("Activate Source Detection to state "+str(state))
        
        for k,v in self.EPNExpButtonDict.items():            
            if "edetectchain" in k:
                if state is False:
                    self.EPNExpButtonDict[k].setChecked(False)
                    self.EPNExpButtonDict[k].setText("yes")
                else:
                    self.EPNExpButtonDict[k].setChecked(True)
                    self.EPNExpButtonDict[k].setText("no")
                    



# ESA (C) 2000-2022
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pymakethumbs.py

from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA

__version__ = f'pymakethumbs (pymakethumbs-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]'



import matplotlib.pyplot as plt
import os
import sys
from pysas.pyutils import pyutils as pyutils
from logger import get_logger
from astropy.wcs import WCS
from astropy.io import fits
import numpy as np
from matplotlib.colors import LogNorm

thumb_log = get_logger('makethumb')

def str_to_bool(string_param):
    '''
    Returns True or False depending on the strings found in parameters that are supposed 
    to be booleans.

    Args:
        string_param: the given parameter.
    
    Output:
        bool
    '''

    if 'Y' in string_param.upper() or 'T' in string_param.upper():
        return True
    else:
        return False


def prepare_text(hdu, manual_parameters, n_image, srcindexstyle, iau = ''):
    '''
    Creates the text for the upper part of the thumbnail.

    Args:
        hdu: the header of the FITS image.
        manual_parameters: tuple containing the information that can be passed via the command line:
         - obsstr
         - inststr
         - erangestr
        n_image: the index number of the image.
        srcindexstyle: the preffix for the index.
        iau: the name from the IAU catalogue. Might be left empty.
    
    Output:
        (plot_text, plot_comments): the text prepared for the plot, and the comments.
    '''

    obsidstr, inststr, erange, commupper, commlower = manual_parameters

    if obsidstr == '' or obsidstr.upper() == 'DEFAULT':
        try:
            obs_id = hdu['OBSID']
        except KeyError:
            obs_id = hdu['OBS_ID']
    else:
        obs_id = obsidstr

    if inststr == '' or inststr.upper() == 'DEFAULT':
        try:
            instr = hdu['INSTRUME']
        except KeyError:
            thumb_log.warning('Could not find the INSTRUME keyword in the given FITS image.')
            instr = 'unknown'
    else:
        instr = inststr

    id_image = int(n_image) + 1
    
    if iau != '':
        id_image = ''
        srcindexstyle = ''

    plot_text = '{0}{1} {2}\n'.format(srcindexstyle, id_image, iau) + \
            '{0} {1} {2}'.format(obs_id, instr, erange)

    plot_comments = commupper + '\n' + commlower

    return (plot_text, plot_comments)


def info_from_catalogue(cat_file, cat_ext, srcnum):
    '''
    Returns the source info from the given catalogue. 

    Args:
        cat_file: the name of the file containing the catalogue.
        cat_ext: the extension in the catalogue in which the data can be found.
        srcnum: the number that identifies the source in the catalogue data.

    Output:
        src_nums: the sources numbers.
        srcids: a list of sources ids.
        detids: the DETID column information.
        numhex: the hexagesimal number for the source number.
        iau_vals: an array containing the IAUNAME data.
    '''

    thumb_log.debug('Getting source info from source catalogue.')

    if srcnum == 0:
        with fits.open(cat_file) as cat_f:
            data_cat = cat_f[cat_ext].data
            try:
                iau_vals = data_cat['IAUNAME']
            except KeyError:
                thumb_log.warning('Could not locate IAUNAME column in the source data catalogue.')
                iau_vals = [''] * len(data_cat)
            try:
                detids = data_cat['detid']
            except KeyError:
                try:
                    detids = data_cat['DETID']
                except KeyError:
                    thumb_log.warning('Could not locate DETID column in catalogue data.')
                    detids = [''] * len(data_cat)
            try:
                srcids = data_cat['srcid']
            except KeyError:
                srcids = [''] * len(data_cat)
            try:
                src_nums = data_cat['src_num']
            except KeyError:
                src_nums = [''] * len(data_cat)
            numhex = []
            for i in src_nums:
                try:
                    n_hex = int(i)
                    numhex.append(hex(n_hex))
                except TypeError:
                    numhex.append('')
    else:
        with fits.open(cat_file) as cat_f:
            data_cat = cat_f[cat_ext].data
            if 'SRC_NUM' in data_cat.names or 'src_num' in data_cat.names:
                iau_vals = data_cat[data_cat['SRC_NUM'] == srcnum]['IAUNAME'][0]
                detids = data_cat[data_cat['SRC_NUM'] == srcnum]['detid'][0]
                srcids = data_cat[data_cat['SRC_NUM'] == srcnum]['SRCID'][0]
                src_nums = data_cat[data_cat['SRC_NUM'] == srcnum]['SRC_NUM'][0]
                try:
                    numhex = hex(int(srcnum))
                except TypeError:
                    thumb_log.warning('Could not convert the given src number to hexagesimal.')
                    numhex = ''
            else:
                try:
                    src_nums = np.array((data_cat['SRC_NUM'][srcnum]))[0]
                except (IndexError, KeyError) as e:
                    thumb_log.warning('Could not find SRC_NUM in the given catalogue table: {}:{}.', cat_file, cat_ext)
                    src_nums = srcnum
                try:
                    iau_vals = np.array((data_cat['IAUNAME'][srcnum]))[0]
                except (IndexError, KeyError) as e:
                    thumb_log.warning('Could not find IAUNAME in the given catalogue table: {}:{}.', cat_file, cat_ext)
                    iau_vals = ''
                try:
                    detids = np.array((data_cat['detid'][srcnum]))[0]
                except (IndexError, KeyError) as e:
                    thumb_log.warning('Could not find DETID in the given catalogue table: {}:{}.', cat_file, cat_ext)
                    detids = ''
                    try:
                        numhex = hex(int(src_num))
                    except TypeError:
                        numhex = ''
                try:
                    srcids = np.array((data_cat['SRCID'][srcnum]))[0]
                except (IndexError, KeyError) as e:
                    thumb_log.warning('Could not find SRCID in the given catalogue table: {}:{}.', cat_file, cat_ext)
                    srcids = ''
                    numhex = ''

    return (src_nums, srcids, detids, numhex, iau_vals)


def source_from_table(src_file, src_ext, srcnum):
    '''
    Returns the coordinates (RA, DEC) and the number of images ethat will be 
    produced by checking the source list.

    Args:
        src_file: the file containing the source list.
        src_ext: the extension in the file in which the data can be found.
        srcnum: the number/index of the source. 0 indicates all the sources available.

    Output:
        coords: a list containing all the values from RA, DEC fund.
        n_images: the number of images that will be produced.
    '''

    thumb_log.debug('Getting source info from source table.')
    
    if srcnum == 0:
        thumb_log.info('Selecting all sources available in file {}:{}.', src_file, src_ext)
        with fits.open(src_file) as src_f:
            try:
                ra_coords = src_f[src_ext].data['RA']
                dec_coords = src_f[src_ext].data['DEC']
                try:
                    src_nums = src_f[src_ext].data['SRC_NUM']
                except KeyError:
                    thumb_log.debug('Could not locate SRC_NUM in file {}. Will try with SRCNUM.', src_file)
                    try:
                        src_nums = src_f[src_ext].data['SRCNUM']
                    except KeyError:
                        src_nums = np.arange(len(ra_coords))
                        thumb_log.warning('Could not locate srcnum info. Using a ordered list instead.')
            except (IndexError, KeyError) as e:
                thumb_log.error('Could not find RA/DEC info in the given source file: {}:{}.', src_file, src_ext)
                sys.exit(0)
    else:
        with fits.open(src_file) as src_f:
            if 'SRC_NUM' in src_f[src_ext].data.names:
                ra_coords = src_f[src_ext].data[src_f[src_ext].data['SRC_NUM'] == srcnum]['RA']
                dec_coords = src_f[src_ext].data[src_f[src_ext].data['SRC_NUM'] == srcnum]['DEC']
                src_nums = srcnum
            else:
                thumb_log.warning('Could not find SRC_NUM in the sourcelist data columns. Selecting the srcnum element in list.')
                try:
                    ra_coords = np.array((src_f[src_ext].data['RA'][srcnum]))
                    dec_coords = np.array((src_f[src_ext].data['DEC'][srcnum]))
                    src_nums = srcnum
                except (IndexError, KeyError) as e:
                    thumb_log.error('Could not find RA/DEC info in the given source file: {}:{}.', src_file, src_ext)
                    sys.exit(0)

    n_images = len(ra_coords)
    if n_images == 0:
        thumb_log.error('No RA/DEC coordinates were found in the SourceList.')
        sys.exit(0)

    coords = []

    for c in range(0, n_images):
        coords.append((ra_coords[c], dec_coords[c]))

    return coords, n_images, src_nums


#################################
############# PLOT ##############
#################################

def plot_thumb(data, hdu, plot_config, plot_text):
    '''
    Makes both the image plot and the corresponding cross.

    Args:
        data: the data to plot as an array.
        hdu: the hdu of the data file.
        plot_config: a tuple containing the parameters that define the plot.
        plot_text: the text to add to the plot.

    Output:
        None
    '''

    thumb_log.info('Running plot_thumb...')

    plt.style.use('dark_background')

    plot_title, colourmap, imagesize, coords, lwidth, sizeratiocross = plot_config
    ra, dec = coords

    wcs = WCS(hdu)
    
    imagesize = imagesize * 3
    
    thumb_log.info('Making plot with center coordinates RA: {} DEC:{}.', ra, dec)

    # defining center and corners
    xc, yc = wcs.wcs_world2pix(ra, dec, 0)
    dx_high, dy_high = wcs.wcs_world2pix(ra + imagesize / 2, dec + imagesize / 6, 0)
    dx_low, dy_low = wcs.wcs_world2pix(ra - imagesize / 2, dec - imagesize / 6, 0)
     
    x_diff = abs(dx_high - dx_low)
    y_diff = abs(dy_high - dy_low)
    
    x_lims = np.array((dx_low, dx_high))
    y_lims = np.array((dy_low, dy_high))


    if x_diff < y_diff:
        dif_lim = x_diff
    else:
        dif_lim = y_diff

    # plotting thumb image

    fig, axs = plt.subplots(2, 1, subplot_kw = {'projection' : wcs})
    fig.set_figheight(10)
    fig.set_figwidth(10)

    axs[1].imshow(data, cmap = colourmap, norm = LogNorm())
    axs[0].axis('off')
    axs[1].axis('off')
    axs[1].set_xlim(x_lims)
    axs[1].set_ylim(y_lims)
    

    # plotting cross
    # getting the center reference
    xlims = axs[1].get_xlim()
    ylims = axs[1].get_ylim()
    
    # set limits to the axis containing the text to match the one from the image
    #axs[0].set_xlim(xlims)
    #axs[0].set_ylim(ylims)

    # invert image:
    axs[1].invert_xaxis()

    # plotting the cross in the image:
    x_diff = abs(x_lims[1] - x_lims[0])
    y_diff = abs(y_lims[1] - y_lims[0])

    # horizontal points
    x = [x_lims[0] - x_diff * 0.25, x_lims[0] - x_diff * (0.45 - sizeratiocross)] 
    y = [yc] * len(x) 
    axs[1].plot(x, y, transform = axs[1].get_transform('pixel'), color = 'C0', linewidth = lwidth)

    x = [x_lims[0] - x_diff * (0.55 + sizeratiocross), x_lims[0] - x_diff * 0.75] 
    y = [yc] * len(x)
    axs[1].plot(x, y, transform = axs[1].get_transform('pixel'), color = 'C0', linewidth = lwidth)

    # vertical points
    y = [y_lims[0] + y_diff * (0.45 - sizeratiocross), y_lims[0] + y_diff * 0.25]  
    x = [xc] * len(y) 
    axs[1].plot(x, y, transform = axs[1].get_transform('pixel'), color = 'C0', linewidth = lwidth)

    y = [y_lims[0] + y_diff * (0.55 + sizeratiocross), y_lims[0] + y_diff * 0.75] 
    x = [xc] * len(y) 
    axs[1].plot(x, y, transform = axs[1].get_transform('pixel'), color = 'C0', linewidth = lwidth)

    # adding text
    plot_text, plot_comments = plot_text # unpacking the text
    axs[0].text(0, 0, plot_text, fontsize = 9)
    axs[0].text(1, 0, plot_comments, fontsize = 9, ha = 'right')


    # fix aspect
    axs[0].set_aspect('equal', adjustable = 'box')
    plt.savefig(plot_title)


    return None


def get_plot_title(gifroot, fnamestyle, gifsuffix, fnameseparator, cat_info = ''):
    '''
    Returns the title for the plot given the values of the parameters.
    
    Args:
        gifroot: the start of the filename.
        fnamestyle: the style of the nomenclature.
        gifsuffix: the final suffix of the filename.
        fnameseparator: the separator to use in the nomenclature.
        cat_info: the information to add to the title from the catalogue. 
    
    Output:
        plot_title: the name of the output file.
    '''
    
    separators = get_separators(fnameseparator)
    use_srcid, use_srcnum, use_detid, use_numhex = get_namestyles(fnamestyle)
    full_str_l = [''] * 4
    full_str = ''

    if not isinstance(cat_info, str):
        src_num, srcid, detid, num_hex = cat_info 
        thumb_log.debug('Catalogue info gathered: {}.', cat_info)
        if use_srcid != -1:
            full_str_l[use_srcid] = use_srcid
        else:
            full_str_l.remove('')
        if use_srcnum != -1:
            full_str_l[use_srcnum] = src_num
        else:
            full_str_l.remove('')
        if use_detid != -1:
            full_str_l[use_detid] = detid
        else:
            full_str_l.remove('')
        if use_numhex != -1:
            full_str_l[use_numhex] = num_hex
        else:
            full_str_l.remove('')
            
    thumb_log.debug('Full string iteration: {}. Separators: {}. full_str: {}', full_str_l, separators, full_str)

    if len(full_str_l) == 0:
        full_str = ''
    elif len(full_str_l) == 1:
        full_str = '{0}{1}'.format(full_str, full_str_l[0])
    else:
        for j in range(0, len(full_str_l)):
            if j == len(full_str_l) - 1: # no need to add separator for the last item
                full_str = full_str + '{0}'.format(full_str_l[j])
            else:
                full_str = full_str + '{0}{1}'.format(full_str_l[j], separators[j])
    
    plot_title = gifroot + full_str + gifsuffix
    
    return plot_title


def get_separators(nameseparators):
    '''
    Returns a list of the characters to be used as separators.

    Args:
        nameseparators: a string parsed from fnameseparators

    Output:
        char_list: a list of four elements containing the characters to be used.
    '''
    
    nameseparators = nameseparators.strip().rstrip()
    char_list = list(nameseparators)

    if len(char_list) == 0:
        thumb_log.warning('Could not find separators. Using _ by default.')
        char_list = ['_'] * 4
    
    while len(char_list) < 4:
        char_list = char_list + [char_list[-1]]

    return char_list


def get_namestyles(fnamestyle):
    '''
    Returns a set of booleans that incorporate info on which information should 
    be used in the naming of the files.

    Args:
        fnamestyle: the line parameter fnamestyle.

    Output:
        use_<param>: a set of four booleans.
    '''

    use_srcid, use_srcnum, use_detid, use_numhex = -1, -1, -1, -1
    fnamestyle = fnamestyle.upper()

    name_styles = fnamestyle.split('_')

    if 'SRCID' in name_styles:
        use_srcid = name_styles.index('SRCID')
    if 'SRCNUM' in name_styles:
        use_srcnum = name_styles.index('SRCNUM')
    if 'DETID' in name_styles:
        use_detid = name_styles.index('DETID')
    if 'NUMHEX' in name_styles:
        use_numhex = name_styles.index('NUMHEX')
   
    return(use_srcid, use_srcnum, use_detid, use_numhex)


################################
########### makethumb ##########
################################

def run(iparsdic):
    '''
    Main function of the script. Will change for 'run'.
    '''

    print(f'Executing {__file__} {iparsdic}')

    # gets the parameters...
    # running parameters
    printparams = iparsdic['printparams']
    printparams = str_to_bool(printparams)
    
    if printparams:
        thumb_log.info('Full parameter info: {}', iparsdic)

    dryrun = iparsdic['dryrun']
    dryrun = str_to_bool(dryrun)
    
    # main data parameters
    imageset = iparsdic['imageset']
    
    if not os.path.isfile(imageset):
        thumb_log.error('The given imageset cannot be found.')
        sys.exit(0)
    else:
        with fits.open(imageset) as imset:
            hdu = imset[0].header
            data = imset[0].data

    # source detection
    withsrclist = iparsdic['withsrclist']
    withsrclist = str_to_bool(withsrclist)
   
    withcoords = iparsdic['withcoords']
    withcoords = str_to_bool(withcoords)

    if not withcoords:
        withsrclist = True

    if not withsrclist and not withcoords:
        thumb_log.error('Either withcoords or withsrclist must be true.')
        sys.exit(0)


    # catalogue
    withcat = iparsdic['withcat']
    withcat = str_to_bool(withcat)

    # source number selection
    if withcat or withsrclist:
        withsrcnum = iparsdic['withsrcnum']
        srcnum = int(iparsdic['srcnum'])


    if withsrclist:
        srclisttab = iparsdic['srclisttab']
        try:
            src_file, src_ext = srclisttab.split(':')
        except ValueError:
            thumb_log.warning('Could not split srclisttab. Will use The first non-primary extension.')
            src_file = srclisttab
            src_ext = 1
        coords, n_images, srcnums = source_from_table(src_file, src_ext, srcnum)
        thumb_log.debug('Info from srclisttab: coords: {}. source numbers: {}.', coords, srcnums)
    else:
        # manual coordinates selection
        if withcoords:
            ra = float(iparsdic['ra'])
            dec = float(iparsdic['dec'])
            n_images = 1
            coords = [(ra, dec)]

    if withcat and not withcoords:
        cattab = iparsdic['cattab']
        try:
            cat_file, cat_ext = cattab.split(':')
        except ValueError:
            thumb_log.warning('Could not split cattab. Will use The first non-primary extension.')
            cat_file = cattab
            cat_ext = 1
        if isinstance(srcnums, int):
            src_nums, srcids, detids, numhex, iau_vals = info_from_catalogue(cat_file, cat_ext, srcnums)
        else:
            src_nums = []
            srcids = []
            detids = []
            numhex = []
            iau_vals = []
            for src in srcnums:
                src_nums_i, srcids_i, detids_i, numhex_i, iau_vals_i = info_from_catalogue(cat_file, cat_ext, src)
                src_nums.append(src_nums_i)
                srcids.append(srcids_i)
                detids.append(detids_i)
                numhex.append(numhex_i)
                iau_vals.append(iau_vals_i)


        thumb_log.debug('Catalogue overall info: src_nums: {}, srcids: {}, detids: {}. Num.Hex: {}.', src_nums, srcids, detids, numhex)

    
    # formatting parameters:
    autofname = iparsdic['autofname']
    autofname = str_to_bool(autofname)
   
    if not autofname:
        outfilename = iparsdic['outfilename']
        plot_title = outfilename
        if n_images != 1:
            thumb_log.warning('Since a fixed filename was passed but more than one images were expected only the first one will be created.')
            n_images = 1
    else:
        gifroot = iparsdic['gifroot']
        fnamestyle = iparsdic['fnamestyle']
        gifsuffix = iparsdic['gifsuffix']
        if '.GIF' in gifsuffix.upper():
            thumb_log.warning('matplotlib does not support GIF format. Changing to png.')
            gifsuffix = '.png'
        fnameseparator = iparsdic['fnameseparator']
        

    srcindexstyle = iparsdic['srcindexstyle']
        
    if srcindexstyle.upper() == 'DEFAULT':
        if withcat:
            srcindexstyle = 'detid'
        elif withsrclist:
            srcindexstyle = 'srcnum'
        else:
            srcindexstyle = 'none'

    if srcindexstyle.upper() == 'NONE':
        srcindexstyle = ''
    elif srcindexstyle.upper() == 'SRCNUM':
        srcindexstyle = '/'
    elif srcindexstyle.upper() == 'HEXSRCNUM':
        srcindexstyle == ''
        thumb_log.warning('srcindexstyle=hexsrcnum is not yet supported. Switched to none')
    elif srcindexstyle.upper() == 'DETID':
        srcindexstyle = ':'
    elif srcindexstyle.upper() == 'SRCID':
        srcindexstyle = '/'
    else:
        srcindexstyle = ''
        thumb_log.warning('Unrecognised format. Using none.')

    # plotting options
    iaunameprefix = iparsdic['iaunameprefix']
    obsidstr = iparsdic['obsidstr']
    erangestr = iparsdic['erangestr']
    inststr = iparsdic['inststr']
    lwidth = iparsdic['lwidth']
    commupper = iparsdic['commupper']
    commlower = iparsdic['commlower']
    imagesize = float(iparsdic['imagesize']) # arcmin
    imagesize = imagesize / 60
    sizeratiocross = float(iparsdic['sizeratiocross'])
    sizeratiocross = sizeratiocross / 100 # the ratio is in %
    colourmap = iparsdic['colourmapid'] # refined for matplotlib
    manual_parameters = (obsidstr, inststr, erangestr, commupper, commlower)

    try:
        colourmap = plt.cm.get_cmap(colourmap)
    except ValueError:
        thumb_log.warning("The selected colourmap is not present in the matplotlib cmap list. Using 'plasma' by default.")
        colourmap = plt.cm.get_cmap('plasma')

    for i in range(0, n_images):
        if autofname:
            if withcat and not withcoords:
                src_num = srcnums[i]
                srcid = srcids[i]
                detid = detids[i]
                num_hex = numhex[i]
                iau = iaunameprefix + iau_vals[i]
                cat_info = src_num, srcid, detid, num_hex
                thumb_log.debug('Plot info for image {}', i)
                plot_title = get_plot_title(gifroot, fnamestyle, gifsuffix, fnameseparator, cat_info)
            else:
                plot_title = gifroot + str(1 + i) + gifsuffix
                thumb_log.debug('Plot title: {}.', plot_title)
        else:
            thumb_log.debug('Plot title: {}.', plot_title)
        
        ra_dec = coords[i]
        
        plot_text = prepare_text(hdu, manual_parameters, i, srcindexstyle)
        plot_parameters = (plot_title, colourmap, imagesize, ra_dec, lwidth, \
                sizeratiocross)
        plot_thumb(data, hdu, plot_parameters, plot_text)
    

    return None



# version.py 
#
# ESA (C) 2000-2021
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#

# pySAS version
VERSION = '2.3.0'

import subprocess, os, sys

def get_sas_version():
    sas_dir = os.environ.get('SAS_DIR')
    if sas_dir == None:
        print('Error: SAS_DIR is undefined')
        sys.exit()
    SAS_RELEASE          = ''
    SAS_AKA              = ''
    SAS_COMPILATION_DATE = ''
    SAS_COMPILATION_HOST = ''
    SAS_COMPILATION_USER = ''
    SAS_PLATFORM         = ''
    SAS_COMMIT_ID        = ''
    cmd = 'sasversion'
    # Start the subprocess
    process = subprocess.Popen(cmd, 
                               bufsize=1,
                               shell=True,
                               text=True,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT,
                               universal_newlines=True)

    # Log stdout and stderr in real-time
    # For non-Python SAS tasks the stout and stderr are combined
    output = []
    for line in process.stdout:
        output.append(f"{line.strip()}")

    # Wait for the process to complete and get the return code
    process.wait()

    for line in output:
        if line.startswith('SAS release'):
            SAS_RELEASE = line.split(':')[1].lstrip()
            if '-' in SAS_RELEASE:
                # If SAS_RELEASE uses dashes (v22>=)
                SAS_AKA = SAS_RELEASE.split('-')[0]
                SAS_COMMIT_ID = SAS_RELEASE.split('-')[1]
            else:
                # If SAS_RELEASE uses underscores (v21<=)
                SAS_AKA = SAS_RELEASE.split('_')[0]
        if line.startswith('Compiled on'):
            SAS_COMPILATION_DATE = line.split(':', 1)[1].lstrip()
        if line.startswith('Compiled by'):
            temp = line.split(':')[1].lstrip()
            try:
                SAS_COMPILATION_USER = temp.split('@')[0]
                SAS_COMPILATION_HOST = temp.split('@')[1]
            except Exception:
                pass
        if line.startswith('Platform'):
            SAS_PLATFORM = line.split(':')[1].lstrip()
    return_list = [SAS_RELEASE,
                   SAS_AKA,
                   SAS_COMPILATION_DATE,
                   SAS_COMPILATION_HOST,
                   SAS_COMPILATION_USER,
                   SAS_PLATFORM,
                   SAS_COMMIT_ID]
    return return_list


# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyrgsimplot.py

from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA

__version__ = f'pyrgsimplot (pyrgsimplot-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 


import logging
import pysas.pyutils.pyutils as pyutils
import os
import numpy as np
import sys
import pysas.pysasplot_utils.pysasplot_utils as sasplt
from astropy.io import fits
import matplotlib.pyplot as plt
from astropy import units as u
import warnings
warnings.filterwarnings('ignore')
from matplotlib.colors import LogNorm
from matplotlib.colors import PowerNorm
import glob
import time
import matplotlib.transforms as tr
from astropy.wcs import WCS
from astropy.coordinates import Angle
from logger import get_logger
PI = u.def_unit('PI', 1 * u.eV)
u.add_enabled_units([PI])

logger = get_logger('pyrgsimplot')


def set_xlabel(xlabel):
    """
    Defines the value for the x label depending on the data taken from the fits'
    header.

    Args:
        xlabel: the original value taken from the fits file.

    Output:
        xlabel: the new value for the x label.
    """

    if xlabel == 'M_TIMES_LAMBDA':
        xlabel = r"$\lambda$" + ' (nm)'
    elif xlabel == 'DETX':
        xlabel = 'DETX (mm)'
    elif xlabel == 'BETA_CORR':
        xlabel = 'DISPERSION ANGLE (rad)'
    elif xlabel == 'LAMBDA':
        xlabel = r"$\lambda$" + r"($\AA$)"
    else:
        pass

    return xlabel


def getobservationdetails(in_dataset):
    """
    Uses pyutils to return the keywords corresponding to the given fits file.

    Args:
        in_dataset: the fits file to evaluate.

    Output:
        fits_info: list containing the exposure, telescope, obs_id, exp_id, instr,
    date_obs, date_end
    """

    logger.debug('Running getobservationdetails...')

    exposure = pyutils.get_key_word(in_dataset, 'EXPOSURE', 0)
    telescope = pyutils.get_key_word(in_dataset, 'TELESCOP', 0)
    obs_id = pyutils.get_key_word(in_dataset, 'OBS_ID', 0)
    exp_id = str(pyutils.get_key_word(in_dataset, 'EXP_ID', 0))[-3:]
    instr = pyutils.get_key_word(in_dataset, 'INSTRUME', 0)
    date_obs = pyutils.get_key_word(in_dataset, 'DATE-OBS', 0)
    date_end = pyutils.get_key_word(in_dataset, 'DATE-END', 0)

    fits_info = [exposure, telescope, obs_id, exp_id, instr, date_obs, date_end]
    
    return fits_info


def set_ylabel(ylabel):
    """
    Defines the value for the y label depending on the data taken from the fits
    header.

    Args:
        ylabel: the original value taken from the fits file.

    Output:
        ylabel: the new value for the y label.
    """

    if ylabel == 'XDSP_CORR':
        ylabel = 'CROSS DISPERSION ANGLE (radians)'
    elif ylabel == 'PHA':
        ylabel = 'PHA (Channel)'
    elif ylabel == 'PI':
        ylabel = 'PI (Channel)'
    elif ylabel == 'PIFUDGE':
        ylabel = 'PI (Channel)'
    elif ylabel == 'DETY':
        ylabel = 'DETY (mm)'
    else:
        pass

    return ylabel


def plot_image(data, hdu, fits_info, xlabel, ylabel, norm, colourmap, region_list, plot_title, kind, srclist):
    """
    Plots an image. Depending on the input, on top of the image the regions from the
    sources might be added.

    Args:
        data: the array to plot.
        hdu: the header from the FITS file to get the WCS reference.
        fits_info: information regarding the file that is being plotted.
        xlabel: name of the xlabel.
        ylabel: name of the ylabel.
        norm: normalization to use.
        colourmap: the colour map of the image.
        region_list: the dictionary containing the regions to be plotted.
        plot_title: the title for the plot.
        kind: the two types of plot to use.
        srclist: file corresponding to the source list to plot the RA and DEC.
    """

    exposure, telescope, obs_id, exp_id, instr, date_obs, date_end, typing = fits_info
   
    if dark_mode:
        plt.style.use('dark_background')

    if norm.upper() == 'LOG':
        norm = LogNorm()
        # Uncomment the following lines to change empty pixels with the minimum value.
        #second_minimum = sorted(set(data.flatten()))[1]
        #data[data == 0] = second_minimum
        norm_name = 'log'
    elif norm.upper() == 'POW':
        norm = PowerNorm(gamma = gamma)
        norm_name = 'power'
    else:
        norm = ''
        norm_name = 'linear'
    
    logger.info('Plotting {}.', plot_title)

    wcs = WCS(hdu)
    fig, ax1 = plt.subplots(subplot_kw = {'projection': wcs})

    try:
        with u.add_enabled_units([PI]):
            if inverted:
                colour_map_parsed = plt.cm.get_cmap(colourmap)
                im = ax1.imshow(data,interpolation = None, origin = 'lower', aspect = 'auto', norm = norm, cmap = colour_map_parsed.reversed())
            else:
                im = ax1.imshow(data,interpolation = None, origin = 'lower', aspect = 'auto', norm = norm, cmap = colourmap)
    except ValueError:
        logger.warning('The input colourmap is not available for matplotlib. Using \'plasma\' by default.')
        colourmap = 'plasma'
        with u.add_enabled_units([PI]):
            if inverted:
                colour_map_parsed = plt.cm.get_cmap(colourmap)
                im = ax1.imshow(data,interpolation = None, origin = 'lower', aspect = 'auto', norm = norm, cmap = colour_map_parsed.reversed())
            else:
                im = ax1.imshow(data,interpolation = None, origin = 'lower', aspect = 'auto', norm = norm, cmap = colourmap)

    # if the plot is reduced to the PI detection even if the sources extend it:
    #ylims = ax1.get_ylim()
    #xlims = ax1.get_xlim()

    if region_list:
        for key in region_list.keys():
            # masking out the non-value coordinates marked as 0:
            filt_valuex = region_list[key][0]
            filt_valuex[filt_valuex == 0] = np.nan
            filt_valuey = region_list[key][1]
            filt_valuey[filt_valuey == 0] = np.nan
            with u.add_enabled_units([PI]):
                if kind == 'endisp':
                    order_key = key[key.upper().find('ORDER') + 6]
                    if not order_key.isnumeric():
                        logger.warning('Could not identify the order in the source key.')
                        colour_plot = 'C0'
                    else:
                        colour_plot = 'C{0}'.format(order_key)
                else:
                    colour_plot = 'C0'
                if 'ext' in key:
                    ax1.plot(filt_valuex, filt_valuey,'--', color = colour_plot, transform = ax1.get_transform('world'))
                else:
                    ax1.plot(filt_valuex, filt_valuey, color = colour_plot, transform = ax1.get_transform('world'))
    
    if srclist:
        text = ''
        s_ra, s_dec, s_label = get_source_details(srclist)
        for i in range(0, len(s_label)):
            aux_text = r"$\bf{Label: }$ " + s_label[i] + r" $\bf{RA: }$ " + s_ra[i] + r" $\bf{DEC: }$ " + s_dec[i] + '\n'
            text = text + aux_text

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    cbar = fig.colorbar(im)

    if norm_name != 'power':
        cbar.ax.set_ylabel('Counts per binned pixel ({} intensity).'.format(norm_name))
    else:
        cbar.ax.set_ylabel('Counts per binned pixel (power (gamma = {}) intensity).'.format(gamma))
    plt.figtext(0.5, 0.01, text)
    #ax1.set_xlim(xlims)
    #ax1.set_ylim(ylims)
    fig.set_size_inches(12, 9)
    ax1.set_title('Date-OBS: {0}  Date-END: {1}\nOBS-ID: {2}  EXP-No: {3}'.format(date_obs, date_end, obs_id, exp_id))
    fig.suptitle('XMM Newton {0} {1} Image'.format(instr, typing), fontsize = 16)
    
    # since more than one file is going to be written, the names must be changed
    if withendispset and withspatialset:
        plot_title = plot_title + kind
    
    try:
        logger.debug('Saving {} {}.', plot_title, out_format)
        plt.savefig(plot_title + '.' + out_format)
    except FileNotFoundError:
        try:
            dirs, rfile = os.path.split(plot_title + '.' + out_format)
            os.makedirs(dirs)
            plt.savefig(plot_title + '.' + out_format)
        except:
            logger.error('Could not create {}. Could not resolve the given path.', plot_title + '.' + out_format)

    
    if device:
        if os.getenv['DISPLAY']:
            plt.show()
        else:
            logger.warning('Display not available.')

    logger.info('Created {}.', plot_title)


def getarrayattributes(data_header):
    """
    From the header of an array, gets the basic information needed.

    Args:
        data_header: the header to extract the data.

    Output:
        (xlabel, ylabel, xref, yref, xbinwidth, ybinwidth, xmin, ymin): tuple
    containing the data extracted.
    """

    logger.debug('Running getarrayattributes...')

    xlabel = data_header['CTYPE1']
    if xlabel == 'M_LAMBDA':
        xlabel = 'LAMBDA'

    ylabel = data_header['CTYPE2']
    xref = float(data_header['CRPIX1'])
    yref = float(data_header['CRPIX2'])
    xbinwidth = float(data_header['CDELT1'])
    ybinwidth = float(data_header['CDELT2'])

    xmin = float(data_header['CRVAL1']) + (1 - xref) * xbinwidth
    ymin = float(data_header['CRVAL2']) + (1 - yref) * ybinwidth
    
    return (xlabel, ylabel, xref, yref, xbinwidth, ybinwidth, xmin, ymin)


def run(iparsdic):
    print(f'Executing {__file__} {iparsdic}')

    t_start = time.time()

    global withspatialset, withendispset

    withspatialset = iparsdic['withspatialset']
    if 'Y' in withspatialset.upper() or 'T' in withspatialset.upper():
        withspatialset = True
    else:
        withspatialset = False

    withendispset = iparsdic['withendispset']
    if 'Y' in withendispset.upper() or 'T' in withendispset.upper():
        withendispset = True
    else:
        withendispset = False
    
    withspatialregionsets = iparsdic['withspatialregionsets']
    if 'Y' in withspatialregionsets.upper() or 'T' in withspatialregionsets.upper():
        withspatialregionsets = True
    else:
        withspatialregionsets = False

    order_list = iparsdic['orderlist']
    order_list = order_list.replace(' ','').split(',')

    global inverted
    inverted = iparsdic['inverted']
    if 'Y' in inverted.upper() or 'T' in inverted.upper():
        inverted = True
    else:
        inverted = False

    global out_format

    withendispregionsets = iparsdic['withendispregionsets']
    if 'Y' in withendispregionsets.upper() or 'T' in withendispregionsets.upper():
        withendispregionsets = True
    else:
        withendispregionsets = False

    global device
    device = iparsdic['device']
    if 'Y' in device.upper() or 'T' in device.upper():
        device = True
    else:
        device = False

    endispset = iparsdic['endispset']
    spatialset = iparsdic['spatialset']

    srclistset = iparsdic['srclistset']
    if not os.path.isfile(srclistset):
        srclistset = None

    plot_title = iparsdic['plotfile']
    plot_title = os.path.abspath(plot_title)
    plot_title, out_format = os.path.splitext(plot_title)
    
    if out_format == '':
        out_format = 'pdf'
        logger.warning('Could not find format. Using {}.{} as default.', plot_title, out_format)
    else:
        out_format = out_format.replace('.', '')
        if not out_format.upper() in ('PDF', 'PNG'):
            logger.error('Format {} not supported. Please use pdf or png.', out_format)
    
    
    colourmap = iparsdic['colour']
    norm = iparsdic['colourmap']
    n_sources = int(iparsdic['srcidlist'])
    
    global gamma
    gamma = float(iparsdic['gamma'])
    
    global dark_mode
    dark_mode = iparsdic['dark_mode']
    if 'Y' in dark_mode.upper() or 'T' in dark_mode.upper():
        dark_mode = True
    else:
        dark_mode = False

    if not withspatialset and not withendispset:
        logger.error('Nothing to plot.')
        sys.exit(0)
    if withspatialset:
        spatialset = iparsdic['spatialset']
    if withendispset:
        endispset = iparsdic['endispset']
        if not os.path.isfile(endispset):
            logger.error('No energy dispersion set provided.')


    ################################
    ######## SPATIAL IMAGE #########
    ################################
    
    logger.info('Starting processing the spatial energy block...')

    if withspatialset:
        try:
            with fits.open(spatialset) as in_dataset:
                fits_info = getobservationdetails(in_dataset)
                fits_info.append('Spatial')
                in_array = in_dataset[0].data
                array_header = in_dataset[0].header
        except FileNotFoundError:
            logger.error('Could not open spatial set file.')
            sys.exit(0)

        dim = in_array.shape
        xlabel, ylabel, xref, yref, xbinwidth, ybinwidth, xmin, ymin = getarrayattributes(array_header)
        xlabel = set_xlabel(xlabel)
        ylabel = set_ylabel(ylabel)
        ymax = ymin + dim[1] * ybinwidth
        xmax = xmin + dim[0] * xbinwidth
        limits = [xmin, ymin, xmax, ymax]

        if withspatialregionsets:
            region_list = collect_regions_spatial(srclistset, n_sources)
        else:
            region_list = None
        
        plot_image(in_array, array_header, fits_info, xlabel, ylabel, norm, colourmap, region_list, plot_title, 'spatial', srclistset)


    ###############################
    ######## ORDERS IMAGE #########
    ###############################

    logger.info('Starting energy dispersion block...')

    if withendispset:
        try:
            with fits.open(endispset) as in_dataset:
                in_array = in_dataset[0].data
                fits_info = getobservationdetails(in_dataset)
                fits_info.append('Energy dispersion')
                array_header = in_dataset[0].header
        except FileNotFoundError:
            logger.error('Could not open the energy dispersion dataset.')
            sys.exit(0)

        dim = in_array.shape
        xlabel, ylabel, xref, yref, xbinwidth, ybinwidth, xmin, ymin = getarrayattributes(array_header)
        xlabel = set_xlabel(xlabel)
        ylabel = set_ylabel(ylabel)
        ymax = ymin + dim[1] * ybinwidth
        xmax = xmin + dim[0] * xbinwidth
        limits = [xmin, xmax, ymin, ymax]

        if withendispregionsets:
            region_list = collect_region_order(srclistset, n_sources, order_list)
        else:
            region_list = None
        plot_image(in_array, array_header, fits_info, xlabel, ylabel, norm, colourmap, region_list, plot_title, 'endisp', srclistset)


    if out_format.upper() == 'PDF' and withendispset and withspatialset:
        pdf_files = glob.glob(plot_title + '*.{}'.format(out_format))
        logger.debug('Merging PDFs... ({}).', pdf_files)
        status = sasplt.merge_pdf(pdf_files, plot_title + '.{}'.format(out_format))
        if status == 0:
            logger.warning('Error while merging the PDFs.')
        else:
            for i in pdf_files:
                os.remove(i)

    t_stop = time.time()
    logger.info('All blocks completed in time {}.', round(t_stop - t_start, 2))


def get_source_details(srclistset):
    """
    Gets information regarding the SRCLIST extension from the given source list.

    Args:
        srclistset: the path to the source list fits file.

    Output:
        tuple containing the list of RA, DEC and labels of the sources.
    """
    
    logger.debug('Evaluating the coordinates of the source list...')

    try:
        with fits.open(srclistset) as f:
            source_table = f[1].data
    except FileNotFoundError:
        logger.error('Could not open the soure list set.')
        sys.exit(0)

    source_ra = []
    source_dec = []
    source_name = []

    for row in source_table:
        source_name.append(row['LABEL'])
        source_ra.append(transform_to_hms(row['RA']))
        source_dec.append(transform_to_dms(row['DEC']))

    return (source_ra, source_dec, source_name)


def transform_to_hms(value):
    """
    Transforms the input degree value into a hour-minute-second tuple format 
    using astropy's angle utility.

    Args:
        value: the raw angle in numeric form.

    Ouput:
        transformed: the tuple with the new data.
    """

    logger.debug('Transforming unit...')
    
    value = str(value)
    transformed = Angle(value + 'd').hms
    str_trans = '{0}H:{1}M:{2}S'.format(int(transformed[0]), int(transformed[1]), round(int(transformed[2]),2))
    
    return str_trans


def transform_to_dms(value):
    """
    Transforms the input degree value into a degree-minute-second tuple format
    using astropy's angle utility.

    Args:
        value: the raw angle in numeric form.

    Ouput:
        transformed: the tuple with the new data.
    """

    logger.debug('Transforming unit...')

    value = str(value)
    transformed = Angle(value + 'd').dms
    str_trans = '{0}D:{1}M:{2}S'.format(int(transformed[0]), int(transformed[1]), round(int(transformed[2]),2))

    return str_trans


def collect_regions_spatial(srclist, n_sources):
    """
    Collects the spatial regions present in the source list, under the source id 
    passed.

    Args:
        srclist: the source list fits file.
        n_sources: the source number that has to be checked.

    Output:
        region_list: a dictionary containing the regions ready to plot.
    """
    
    logger.debug('Collecting the spatial regions...')

    if n_sources == 0:
        return None

    region_list = dict()
    ext = ''
    sp_regions = []
    
    try:
        with fits.open(srclist) as f:
            for h in f:
                if 'SPATIAL' in h.name:
                    sp_regions.append(h.name)
            if len(sp_regions) == 1:
                ext = sp_regions[0]
            elif len(sp_regions) >= 1:
                for i in sp_regions:
                    if '_SRC{}'.format(n_sources) in i:
                        ext = i
            if ext == '':
                logger.warning('Could not locate spatial regions in the source list.')
                return None
            else:
                dimen, xtag, ytag, component = f[ext].data.names
                for i in range(0, len(f[ext].data)):
                    if '!' in f[ext].data[dimen][i]:
                        region_list.update({'ext{}'.format(i) : (f[ext].data[i][xtag], f[ext].data[i][ytag])})
                    else:
                        region_list.update({'inc{}'.format(i) : (f[ext].data[i][xtag], f[ext].data[i][ytag])})
    except FileNotFoundError:
        logger.error('Could not open the source list file.')
        sys.exit(0)

    logger.debug('Spatial regions collected.')

    return region_list


def collect_region_order(srclist, n_sources, order_list):
    """
    Returns a dictionary containing all the region plots for the given source list, the 
    specified order/orders and the specified source.

    Args:
        srclist: the path to the file for the source list.
        n_sources: the source id to find in the source list.
        order_list: the energy orders to be evaluated.

    Output:
        region_list: the dictionary containing the regions.
    """

    logger.debug('Selecting regions for energy dispersion plot...')

    if n_sources == 0:
        return None

    region_list = dict()
    ext = []
    en_regions = []

    try:
        with fits.open(srclist) as f:
            for h in f:
                if 'ORDER' in h.name:
                    en_regions.append(h.name)
            if len(en_regions) == 1:
                ext = en_regions[0]
            elif len(en_regions) >= 1:
                for i in en_regions:
                    if '_SRC{}'.format(n_sources) in i:
                        for order in order_list:
                            if '_{}'.format(order) in i:
                                ext.append(i)
            if len(ext) == 0:
                logger.warning('WARNING: no regions were matched for the srcidlist.')
                return None
            else:
                for e in ext:
                    dimen, xtag, ytag, component = f[e].data.names
                    for i in range(0, len(f[e].data)):
                        if '!' in f[e].data[dimen][i]:
                            region_list.update({'ext{}'.format(e + ':' + str(i)) : (f[e].data[i][xtag], f[e].data[i][ytag])})
                        else:
                            region_list.update({'inc{}'.format(e + ':' + str(i)) : (f[e].data[i][xtag], f[e].data[i][ytag])})
    except FileNotFoundError:
        logger.error('Could not open the source list file.')
        sys.exit(0)

    logger.debug('Energy dispersion regions collected.')

    return region_list



# ESA (C) 2000-2021
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# error.py
#
"""error.py

The class Error is an interface to the SAS task saserrstr
 
Instances of class Error, can have the following parameters:

client     : The task that produces the error
msgLayer   : The task layer where the error is originated. 
             See below for the keys in dict msgLayer.
msgLevel   : The importance level of the message. See below for the keys in dict msgLevel.
code       : The code of the message: warning, error, fatal.
msg        : The error message itself.
outMessage : The output message.

Below several possible use cases of class Error for different arguments:

Error( client=name, msgLayer='AppMsg', msgLevel='SilentMsg', msg='Message from the SAS (silent)' ).message()

Error( client=name, msgLayer='AppMsg', msgLevel='SparseMsg', msg='Message from the SAS (sparse)' ).message()

Error( client=name, msgLayer='AppMsg', msgLevel='VerboseMsg', msg='Message from the SAS (verbose)' ).message()

Error( client=name, msgLayer='AppMsg', msgLevel='NoisyMsg', msg='Message from the SAS (noisy)' ).message()

print('-------------------------------------------------------------')

Error( client=name, code='warningCode', msg='Warning message' ).warning()

Error( client=name, code='errorCode', msg='Error message' ).error()

Error( client=name, code='fatalCode', msg='Fatal message' ).fatal()
"""

# Standard library imports
import sys, subprocess

# Third party imports

# Local application imports


class Error:
    """Error Class definitions"""

    msgLayerValues = {
        'UIMsg'     : '1',
        'MetaMsg'   : '2',
        'AppMsg'    : '3',
        'AppLibMsg' : '4',
        'LibMsg'    : '5',
        'CoreMsg'   : '6',
        'KernelMsg' : '7',
        }

    msgLevelValues = {
        'SilentMsg'    : '0',
        'SparseMsg'    : '1',
        'VerboseMsg'   : '2',
        'NoisyMsg'     : '3',
        }

    def __init__(self, client=None, msgLayer=None, msgLevel=None, code=None, msg=None, outMessage=None):
        self.client         = client
        if msgLayer:
           self.msgLayer    = self.msgLayerValues[msgLayer]
        if msgLevel:
           self.msgLevel    = self.msgLevelValues[msgLevel]
        self.code           = code
        self.msg            = msg
        self.outMessage     = outMessage

    def message(self):
        opt = '-m'
        outMessage = ['saserrstr' , opt , self.client, self.msgLayer , self.msgLevel , self.msg ]
        # Spawn 'outMessage' command to the shell through 'saserrstr'
        returncode = subprocess.call(outMessage) 
        if(returncode):
          self.code = "InternalError" 
          self.msg  = "SAS::error::warning failed" 
          self.fatal()
	
    def warning(self):
        opt = '-w'
        outMessage = ['saserrstr', opt, self.client, self.code, self.msg ]
        # Spawn 'outMessage' command to the shell through 'saserrstr'
        returncode = subprocess.call(outMessage)
        if(returncode):
          self.code = "InternalError" 
          self.msg  = "SAS::error::warning failed" 
          self.fatal()
	
    def error(self):
        opt = '-e'
        outMessage = ['saserrstr', opt, self.client, self.code, self.msg ]
        # Spawn 'outMessage' command to the shell through 'saserrstr'
        returncode = subprocess.call(outMessage)
        sys.exit(1)

    def fatal(self):
        opt = '-f'
        outMessage = ['saserrstr', opt, self.client, self.code, self.msg ]
        # Spawn 'outMessage' command to the shell through 'saserrstr'
        returncode = subprocess.call(outMessage)
        sys.exit(1)



# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pysasplot_utils.py

import os, numbers, sys, re, warnings, pickle
import pysas.pyutils.pyutils as pyutils
from astropy.io import fits
from astropy.wcs import WCS
import matplotlib.pyplot as plt
from astropy.table import Table
import numpy as np
import matplotlib.backends.backend_pdf
from pypdf import PdfMerger
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection
from matplotlib.ticker import StrMethodFormatter
import xspec

def plot_spectra_model(spectrum,plot_file_name='spectra_model_plot.png'):
    """
    Convenient function to plot a spectrum and XSPEC model.

    Inputs:
        spectrum - pyXSPEC Spectrum object created using xspec.Spectrum.
                   Spectrum object must contain at least one model 
                   created using xspec.Model.
        plot_file_name - File name of the plot.
    
    Returns:
        Matplotlib figure object and two axis objects.
    """
    xspec.Plot.device='/null'
    xspec.Plot.xAxis = 'keV'

    # Pull off data for main plot
    xspec.Plot('data')
    energy = xspec.Plot.x()
    counts = xspec.Plot.y()
    folded = xspec.Plot.model()
    xErrs = xspec.Plot.xErr()
    yErrs = xspec.Plot.yErr()

    # Pull off data for ratio plot
    xspec.Plot('ratio')
    ratio = xspec.Plot.y()
    r_xerror = xspec.Plot.xErr()
    r_yerror = xspec.Plot.yErr()

    # Get bin edges for "stairs" plot
    bin_edges = []
    for i in spectrum.energies: bin_edges.append(i[0])
    bin_edges.append(spectrum.energies[-1][1])

    # Make the figure and two subplots
    fig, (ax0, ax1) = plt.subplots(nrows=2, sharex=True, height_ratios=[2.5, 1],figsize=(9, 7))

    # Main plot
    ax0.errorbar(energy, counts, yerr=yErrs, xerr=xErrs, linestyle='', marker='')
    ax0.stairs(folded,bin_edges, color='r')
    ax0.set_xscale('log')
    ax0.set_yscale('log')
    ax0.set_xlim([bin_edges[0], bin_edges[-1]])
    ax0.tick_params(top=True,axis="x",direction="in",which='both')
    ax0.tick_params(axis="y",direction="in",which='both',right=True)
    ax0.set_ylabel('counts sec$^{-1}$ keV$^{-1}$')
    ax0.set_title('Data and Folded Model')

    # Ratio plot
    ax1.errorbar(energy, ratio, yerr=r_yerror, xerr=r_xerror, linestyle='', marker='')
    ax1.axhline(y=1, color='g')
    ax1.set_xscale('log')
    ax1.tick_params(top=True,axis="x",direction="in",which='both')
    ax1.tick_params(axis="y",direction="in",which='both')
    ax1.xaxis.set_major_formatter(StrMethodFormatter('{x:.1f}'))
    ax1.xaxis.set_minor_formatter(StrMethodFormatter('{x:.1f}'))
    ax1.set_xlabel('Energy (keV)')
    ax1.set_ylabel('Ratio')

    # This puts the plots together with no space in between
    plt.subplots_adjust(hspace=.0)

    # Save plot to file
    fig.savefig(plot_file_name)

    return fig, ax0, ax1

def quick_image_plot(image_file,
                     xlabel = None,
                     ylabel = None,
                     title  = None,
                     vmin   = 1.0,
                     vmax   = 1e2,
                     save_file = False,
                     out_fname = 'image.png'):
    """
    Displays a FITS image file. Returns the axis handle.
    """
    hdu = fits.open(image_file)[0]
    wcs = WCS(hdu.header)

    ax = plt.subplot(projection=wcs)
    plt.imshow(hdu.data, origin='lower', norm='log', vmin=vmin, vmax=vmax)
    ax.set_facecolor("black")
    plt.grid(color='blue', ls='solid')
    if not xlabel is None:
        plt.xlabel(xlabel)
    if not ylabel is None:
        plt.ylabel(ylabel)
    if title is None:
        plt.title(f'FITS Image')
    else:
        plt.title(title)
    
    plt.colorbar()
    plt.show()
    if save_file:
        plt.savefig(out_fname)

    return ax

def quick_light_curve_plot(light_curve_file,
                           tstart = None,
                           tend   = None,
                           title  = None,
                           save_file = False,
                           out_fname = 'light_curve.png'):
    """
    Plots a light curve from a light curve FITS file.
    """
    ts = Table.read(light_curve_file,hdu=1)
    plt.plot(ts['TIME'],ts['RATE'])
    plt.xlabel('Time (s)')
    plt.ylabel('Count Rate (ct/s)')
    if tstart is not None:
        plt.xlim(left=tstart)
    if tend is not None:
        plt.xlim(right=tend)
    if title is not None:
        plt.title(title)
    plt.show()
    if save_file:
        plt.savefig(out_fname)

def text_plot(fits_file, extra_text = ''):
    """
    Prepares the text for the plot reading Keywords from the given FITS file.

    Args:
        fits_file: the path to the FITS file.
        extra_text: additional text to add to the plot. Empty by default.

    Output:
        text_plot: the text as a string to be added to the plot.
    """

    obs = pyutils.get_key_word(fits_file, 'OBS_ID')
    expo = pyutils.get_key_word(fits_file, 'EXPIDSTR')
    inst = pyutils.get_key_word(fits_file, 'INSTRUME')
    revolut = pyutils.get_key_word(fits_file, 'REVOLUT')
    ra = pyutils.get_key_word(fits_file, 'RA_OBJ')
    dec = pyutils.get_key_word(fits_file, 'DEC_OBJ')
    obj = pyutils.get_key_word(fits_file, 'OBJECT')
    date_obs = pyutils.get_key_word(fits_file, 'DATE-OBS')

    text_plot = 'OBS ID: {}\nInstrument: {}\nExposure: {}\nRevolution: {}\nDate: {}\nRA & DEC: ({} {})\n\n{}'.format(obs, inst, expo, revolut, date_obs, ra, dec, extra_text)

    return text_plot

def ingest_data(fits_file, in_data, label = 'label'):
    """
    Prepares the data from the input in a key-value, ready to plot with the
    other functions.

    Args:
        fits_file: the path or the HDU mobject for the fits file.
        in_data: the data to use. Can by an ndarray object or string-like, as in extension:card.
        label: the label to address the data.

    Output:
        data: the adapted dictionary with the {label:data}
    """

    if isinstance(fits_file, str):
        try:
            f = fits.open(fits_file)
        except (FileNotFoundError, ValueError) as e:
            warnings.warn('Could not locate file {}. Quitting...'.format(fits_file))
            sys.exit(0)
    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        f = fits_file
    else:
        warnings.warn('Cannot open the FITS file. Checking the card for array instead of extensions...')

    data = dict()

    if isinstance(in_data, np.ndarray):
        print('Input data: numpy array.')
        data.update({label : in_data})
    elif isinstance(in_data, str):
        in_data = in_data.strip()
        ext, card = in_data.split(':')
        try:
            data.update({ext + ':' + card : f[ext].data[card]})
        except KeyError:
            print('Could not find the combination of extension-key {} in the current FITS file.'.format(ext + ':' + card))
            f.close()
            return None
    elif is_iterable(in_data):
        for comb in range(0, len(in_data)):
            if isinstance(in_data[comb], np.ndarray):
                if is_iterable(label):
                    try:
                        data.update({label[comb] : in_data[comb]})
                    except IndexError:
                        data.update({'{0}_{1}'.format(label[-1], comb) : in_data[comb]})
                else:
                    data.update({'{0}_{1}'.format(label, comb) : in_data[comb]})
            else:
                try:
                    in_data[comb] = in_data[comb].strip()
                    ext, card = in_data[comb].split(':')
                    try:
                        data.update({ext + ':' + card : f[ext].data[card]})
                    except KeyError:
                        print('Could not find the following combination key: {}.'.format(ext + ':' + card))
                        continue
                except KeyError:
                    print('Could not find combination {} in the given FITS file.'.format(ext + ':' + card))
                    continue
    else:
        print('Format unrecognised for x data. Quitting...')
        f.close()
        return None
    
    f.close()

    return data

def is_iterable(py_obj):
    """
    Given a Python object, will return True or False if it's and iterable object aside
    from string.

    Args:
        py_obj: the object that has to be evaluated.

    Output:
        bool: whether or not is iterable.
    """

    from collections.abc import Iterable

    if isinstance(py_obj, Iterable) and not isinstance(py_obj, str):
        return True
    else:
        return False

def get_time_deltas(fits_file, extension = 1, points = 512, x_data = 'TIME', card = 'TIMEDEL'):
    """
    Finds the proper Delta intervals for a FITS file.

    Args:
        fits_file: the fits file: either the path or one already opened.
        extension: the extension for the time sereis to be evaluated. 0 by
    default (primary HDU).
        points: the final number of points wanted. Default: 512.
        x_data: the data to evaluate the x-axis delta. TIME by default.
        card: the card to search in the extension's header. TIMEDEL by default.

    Ouput:
        (DTin, DTout): the initial and final delta time intervals.
        Will return (1, 1) if the file could not be opened or if the extension
    does not contain a proper TIME data column (implying no rebinning).
    """
    
    if not isinstance(fits_file, Table):
        timedel = pyutils.get_key_word(fits_file, card, extension)
    else:
        timedel = 'unknown'

    if isinstance(fits_file, str):
        try:
            with fits.open(fits_file) as f:
                t0 = f[extension].data[x_data][0]
                tf = f[extension].data[x_data][-1]
                fsize = f[extension].data.size
        except FileNotFoundError:
            print('Could not locate file.')
            return (1,1)
    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        try:
            t0 = fits_file[extension].data[x_data][0]
            tf = fits_file[extension].data[x_data][-1]
            fsize = fits_file[extension].data.size
        except KeyError:
            print('Could not locate TIME column in data at [{0}].'.format(extension))
            return (1, 1)
    elif isinstance(fits_file, Table):
        t0 = fits_file[x_data][0]
        tf = fits_file[x_data][-1]
        fsize = len(fits_file)

    if timedel == 'unknown':
        DTin = int(np.round(abs(tf - t0) / fsize))
    else:
        DTin = timedel

    DTout = int(np.round((tf - t0) / points))

    return DTin, DTout

def TSrebin(tstable, DTin, DTout, t_xlabel = 'TIME', t_ylabel = 'RATE', t_elabel = 'ERROR', fracexp = 'FRACEXP'):
    """
    Re-binning function.

    Input:
        tstable: time series with TIME, RATE and ERROR columns.
        DTin: the input bin size.
        DTout: the output bin size.
        t_xlabel: the label to search for the x value. TIME by default.
        t_ylabel: the label to search for the y value. RATE by default.
        t_elabel: the label to search for the error values. ERROR by default.
        fracexp: the label to search for the FRACEXP. FRACEXP by default.

    Output:
        ntstable: a Table object with the corrected bin size.
    """

    #
    # Check input Table Columns, in case it's been loaded as a rec array (as in ext.data)
    #

    if isinstance(tstable, fits.fitsrec.FITS_rec):
        tstable = Table(tstable)

    colNames = tstable.keys() # Input ColumnNames
    mandatoryColumns = [t_xlabel, t_ylabel, t_elabel, fracexp] # Mandatory ColumnNames
    diff = set(mandatoryColumns) - set(colNames)

    fracexp_chk = 0
    if diff == {fracexp}:
        print('FRACEXP is missing. fracexp = 1')
        fracexp_chk = 1
    elif len(diff) != 0:
        print('Missing Columns in input Table : ', diff, '\n Exiting.')
        sys.exit(1)

    #
    # Re-Binning of the Table 'tstable'
    #

    tstable.sort(t_xlabel)
    ntime = tstable[t_xlabel]
    nrate = np.nan_to_num(tstable[t_ylabel])
    nerror = np.nan_to_num(tstable[t_elabel])

    if fracexp_chk == 1:
        fracexp = np.linspace(1, 1, len(ntime))
    else:
        fracexp = np.nan_to_num(tstable[fracexp])

    tcum = np.nancumsum(tstable[t_xlabel])
    rcum = np.nancumsum(nrate * fracexp) # Errors propagation are weighted by the FRACEXP from the TS
    ecum = np.nancumsum(nerror**2 * fracexp)
    fcum = np.nancumsum(fracexp)

    nin = len(rcum) # Number of In points

    t_bin_frac = int(np.round(DTout / DTin))

    if t_bin_frac <= 1: # No re-binning
        ntstable = tstable
        new_time_bin = t_bin_frac
        print('    No re-binning needed for this Light Curve. Input / Output time bins ratio : ', t_bin_frac)
    else:
        nout = int(nin/t_bin_frac) # Number of Out points. t_bin_frac == DTout / DTin ('TIMEDEL'). Re-binning factor

        #print('    Input / Output time bins / ratio : ', DTin, 'sec', ' / ',DTout, 'sec', ' / ', t_bin_frac)

        #print('    Input / Output dots number       : ', nin, ' / ',nout)

        i1 = np.arange(nout) * t_bin_frac
        i2 = np.concatenate((i1[1:] - 1, np.array([nin - 1])))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = RuntimeWarning) #
            if fracexp_chk == 0: # TS has FRACEXP column => rate and error are weighted by FRACEXP
                nrate  = ((rcum[i2] - rcum[i1]) / (fcum[i2] - fcum[i1])    )
                nerror = (np.sqrt(np.nan_to_num(ecum[i2] - ecum[i1])) / (fcum[i2] - fcum[i1]))
                ntime = (tstable[t_xlabel][i2] + tstable[t_xlabel][i1]) / 2.
                nfrac = (fcum[i2] - fcum[i1]) / (i2 - i1)
                ntstable = Table([ntime, nrate, nerror, nfrac], names = [t_xlabel, t_ylabel, t_elabel, fracexp])
            else: # TS has NOT FRACEXP column => fracexp = 1
                nrate  = ((rcum[i2]-rcum[i1]) / (fcum[i2] - fcum[i1])    )
                nerror = (np.sqrt(np.nan_to_num(ecum[i2] - ecum[i1]))/(fcum[i2]-fcum[i1]))
                ntime = (tstable[t_xlabel][i2] + tstable[t_xlabel][i1])/2.
                ntstable = Table([ntime, nrate, nerror], names = [t_xlabel, t_ylabel, t_elabel])

    return ntstable

def load_figures(list_of_saved_figs):
    """
    Loads a list of pickled figures (passed as a list of paths) and returns a list of figure objects.

    Args:
        list_of_saved_figs: a list containing the paths of the pickled figures.

    Output:
        saved_figs: a list of figures.
    """

    saved_figs = []

    for i in list_of_saved_figs:
        with open(i, 'rb') as f:
            try:
                unpickled = pickle.load(f)
                saved_figs.append(unpickled)
            except AttributeError:
                warnings.warn('Could not open the pickled file: callbacks error.')
                saved_figs.append(None)

    return saved_figs

def saves_in_pdf(fig_list, output_name = 'output', papertype = 'a4'):
    """
    Saves the given figures into a single pdf file.

    Args:
        fig_list: a list or tuple containing the Matplotlib elements to be
    saved into a pdf. Recommended: matplotlib.figure.Figure.
        output_name: the name for the pdf file.
        papertype: the paper type for the output. A4 by default.

    Output:
        the absolute path of the created pdf.
        None if the list is empty or the type of fig_list is not correct.
    """

    from matplotlib.backends.backend_pdf import PdfPages

    try:
        if len(fig_list) == 0:
            warnings.warn('The given list is empty.')
            return None
    except TypeError:
        warnings.warn('The input fig_list is not iterable. Quitting...')
        return None

    missed_fig = 0

    with PdfPages('{}'.format(output_name)) as pdf:
        for i in range(0, len(fig_list)):
            try:
                pdf.savefig(fig_list[i], papertype = papertype)
            except ValueError:
                warnings.warn('Index {0} cannot be saved as it is not recognised as a Figure. Skipping...'.format(i))
                missed_fig = missed_fig + 1
                continue

    if missed_fig == len(fig_list):
        warnings.warn('The pdf is empty due to all figures being non-compatible.')

    return os.path.abspath('{}'.format(output_name))

def simple_1d_hist(y_data, x_label = 'x', y_label = 'y', add_text = '', scale = '', nbins = 20, sharebins = False, \
    fits_file = '', fits_info = True, plot_title = '', outformat = 'pdf'):
    """
    Utility for simple 1 dimention histogram.

    Args:
        y_data: the data for the histogram. Uses the inputs from ingest_data. For example, if fits_file has an input: y_data = ['RATE:RATE', array1, array2...].
        x_label: the label for the x-axis.
        y_label: the label for the y-axis.
        add_text: additional text to add alongside the histogram.
        scale: the desired scale to adapt the y-data.
        nbins: the number of bins for the histogram. 20 by default.
        sharebins: whether or not force all the data to have the same bins distribution.
        fits_file = the fits file to use (if any).
        fits_info: Bool. Whether or not show basic info from the FITS file.
        plot_title: the title of the plot.
        outformat: the format (according to matplotlib. PDF by default.
    """

    plt.style.use('ggplot')
    y_data = ingest_data(fits_file, y_data, y_label)

#    if y_data_errors != '':
#        y_data_errors = ingest_data(fits_file, y_data_errors, y_label)

    fig = plt.figure(figsize = (16, 9), facecolor = 'white')

    if fits_info or add_text != '':
        ax = fig.add_axes([0.3, 0.3, 0.6, 0.4])
    else:
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    if sharebins:
        y_data_list = list(y_data.values())
        nbins = np.histogram(np.hstack(y_data_list), bins = nbins)[1]

    for key in y_data.keys():
        try:
            #y_key_error = y_data_errors[key]
            y_error_flag = True
        except (KeyError, TypeError) as e:
            y_error_flag = False
        if not y_error_flag:
            heights, bins, _ = ax.hist(y_data[key], label = str(key), histtype = 'step', fill = True, alpha = 0.2, bins = nbins)
        else:
            heights, bins, _ = ax.hist(y_data[key], label = str(key), histtype = 'step', fill = True, alpha = 0.2, bins = nbins)
            #TODO add errors as separate plotting

    ax.legend(bbox_to_anchor = (1.05, 1), loc = 'upper left', borderaxespad = 0.)

    if fits_file != '' and fits_info:
        add_text = text_plot(fits_file, add_text)

    if add_text != '':
        plt.gcf().text(0.02, 0.5, add_text, fontsize = 14)
    plt.subplots_adjust(top = 0.3)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    if is_iterable(x_label):
        x_label = x_label[0]
    if is_iterable(y_label):
        y_label = y_label[0]

    if plot_title == '':
        plot_title = '{0}_{1}_hist'.format(x_label, y_label)

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(plot_title)

    # writes the filename filtering out non_alphabetical characters
    file_name = re.sub(r'\W+', '', plot_title)

    # will save the plot. If the program is running in a Notebook environment, will show the plot.
    if pyutils.is_notebook():
        plt.show()
        if outformat.upper() == 'PDF':
            saves_in_pdf([fig], file_name)
        elif outformat.upper() == 'PNG':
            plt.savefig(file_name + '.png')
        else:
            warnings.warn('Unrecognised format, will not be saved.')
    else:
        if outformat.upper() == 'PDF':
            saves_in_pdf([fig], file_name)
        else:
            try:
                plt.savefig(file_name + '.{}'.format(outformat))
            except ValueError:
                warnings.warn('Unrecognised format, will not be saved.')

    return fig

def simple_2d_plot(x_data, y_data, x_data_errors = '', y_data_errors = '', plot_title = '', x_label = 'x', y_label = 'y', fits_file = '', fits_info = True, \
    add_text = '', scale = '', points = 512, outformat = 'png'):
    """
    Utility for simple 2 dimentional plots. The data can be:
        from FITS files, written as in EXT:CARD.
        iterable types, including numpy arrays.
        Iterable types with a mix of the above.

    If using a FITS file, always use the phrasing EXT:CARD to refer to an
    extension and card data. Introduce first this information and leave the
    other type of arrays for last.

    Args:
        x_data: the x coordiantes of the data.
        y_data: the y coordinates of the data.
        x_data_errors: errors for the x axis.
        y_data_errors: errors for the y axis.
        plot_title: title for the plot. Will be used also when writting the file.
        x_label: label for the x axis. x by default.
        y_label: label for the y axis. y by default.
        fits_file: only needed if requesting an extension:card in one of the data inputs.
        add_text: additional text to add. None by default.
        scale: the scale to apply.
        points: the number of points. Required in case of rebinning.
        outformat: the format for the output.
    """

    plt.style.use('ggplot')


    if isinstance(fits_file, Table):
        x_data, y_data, x_data_errors, y_data_errors = ingest_table(fits_file, points, x_data, y_data, x_error_label = x_data_errors, y_errors_label = y_data_errors)
        fits_info = False # it's impossible to extract those values from a table object
    else:
        x_data = ingest_data(fits_file, x_data, x_label)
        y_data = ingest_data(fits_file, y_data, y_label)

        if y_data_errors != '':
            y_data_errors = ingest_data(fits_file, y_data_errors, y_label)
        if x_data_errors != '':
            x_data_errors = ingest_data(fits_file, x_data_errors, x_label)

    fig = plt.figure(figsize = (16, 9), facecolor = 'white')
    if fits_info or add_text != '':
        ax = fig.add_axes([0.3, 0.3, 0.6, 0.4])
    else:
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    if len(list(x_data.keys())) == 1:
        x_data_key = list(x_data.keys())[0]
        x_data = list(x_data.values())[0]

        for key in y_data.keys():
            try:
                y_key_error = y_data_errors[key]
                y_error_flag = True
            except (KeyError, TypeError) as e:
                y_error_flag = False
            try:
                x_key_error = x_data_errors[key]
                x_error_flag = True
            except (KeyError, TypeError) as e:
                x_error_flag = False
            if not y_error_flag and not x_error_flag:
                try:
                    ax.plot(x_data, y_data[key], marker='.', label = str(key), linewidth = 0.8)
                    ax.legend(bbox_to_anchor = (1.05, 1), loc = 'upper left', borderaxespad=0.)
                except ValueError:
                    print('Mismatching sizes for x and y: {} and {}. Omitting this curve.'.format(len(x_data), y_data[key].size))
                    continue
            elif x_error_flag and not y_error_flag:
                try:
                    ax.errorbar(x_data, y_data[key], xerr = x_key_error, marker='.', label = str(key), linewidth = 0.8)
                except ValueError:
                    print('Mismatching between the given arrays. Key: {}.'.format(key))
                    continue
            elif y_error_flag and not x_error_flag:
                try:
                    ax.errorbar(x_data, y_data[key], yerr = y_key_error, marker='.', label = str(key), linewidth = 0.8)
                except ValueError:
                    print('Mismatching between the given arrays. Key: {}.'.format(key))
                    continue
            elif y_error_flag and x_error_flag:
                try:
                    ax.errorbar(x_data, y_data[key], xerr = x_key_error, yerr = y_key_error, marker='.', label = str(key), linewidth = 0.8)
                except ValueError:
                    print('Mismatching between the given arrays. Key: {}.'.format(key))
                    continue

    if fits_info and (isinstance(fits_file, str) or isinstance(fits_file, fits.hdu.hdulist.HDUList)):
        add_text = text_plot(fits_file, add_text)

    plt.gcf().text(0.02, 0.5, add_text, fontsize=14)
    #plt.subplots_adjust(top=0.3)

    if scale != '':
        try:
            plt.yscale(scale)
        except ValueError:
            warnings.warn('The given scale is not supported. Using linear scale instead.')
            plt.yscale('linear')

    if is_iterable(x_label):
        x_label = x_label[0]
    if is_iterable(y_label):
        y_label = y_label[0]

    if plot_title == '':
        plot_title = '{0}_{1}_plot'.format(x_label, y_label)

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(plot_title)

    file_name = re.sub(r'\W+', '', plot_title)
    # will save the plot. If the program is running in a Notebook environment, will show the plot.
    if pyutils.is_notebook():
        plt.show()
        if outformat.upper() == 'pdf':
            saves_in_pdf([fig], file_name)
        else:
            plt.savefig(file_name + '.{}'.format(outformat))
    else:
        if outformat.upper() == 'pdf':
            saves_in_pdf([fig], file_name)
        else:
            try:
                plt.savefig(file_name + '.{}'.format(outformat))
            except ValueError:
                warnings.warn('Unrecognised format. Will not save plot.')

    return fig

def ingest_table(table, points, x_label, y_labels, x_error_label = None, y_errors_label = None):
    """
    Ingest the data from a given table and returns the data using Python
    dictionaries with labels. Only labels from the table are supported.

    Args:
        table: the table object.
        points: the number of points wanted. A rebinning will be applied to the table.
        x_label: the label for the x axis.
        y_labels: the label (or array of labels) for the y data.
        x_error_label: the label for the error for the x-coordinates.
        y_error_labels: the label (or array of labels) for the error of the
    y-coordinates.

    Output:
        (x_data, y_data, x_error, y_data_errors): set of dictionaries containing
    the information from the table.
    """

    DTin, DTout = get_time_deltas(table, 1, points)
    t_bin_frac = int(np.round(DTout / DTin))
    if t_bin_frac <= 1:
        print('No rebinning.')
    else:
        table = TSrebin(table, DTin, DTout)

    x_data = {x_label : table[x_label]}
    y_data = dict()

    if is_iterable(y_labels):
        for label in y_labels:
            y_data.update({label : table[label]})
    elif isinstance(y_labels, str):
        y_data.update({y_labels : table[y_labels]})

    nrows = len(table)
    x_error = dict()

    try:
        if x_error_label:
            x_error = x_error.update({x_error_label : table[x_error_label]})
    except KeyError:
        print('Could not find errors for the x axis in {0}.'.format(x_error_label))

    y_error = dict()

    if y_errors_label:
        if is_iterable(y_errors_label):
            for label in y_errors_label:
                try:
                    y_error.update({label : table[label]})
                except KeyError:
                    print('Could not find {0} in the given table.'.format(label))
                    continue
        elif isinstance(y_errors_label, str):
            try:
                y_error.update({y_labels : table[y_labels]})
            except KeyError:
                print('Could not find {0} in the given table.'.format(y_labels))

    return(x_data, y_data, x_error, y_error)

def merge_pdf(pdf_files, output_file):
    """
    Merges several PDF files passed as the arguments. 
    
    Args:
        pdf_files: list or tuple containing the paths to the pdfs.
        output_file: the name of output file.

    Output:
        1 depending on whether or not the process has been completed; 0 if not.
    """

    merger = PdfMerger()
    total_files = len(pdf_files)
    e_count = 0

    for pdf in pdf_files:
        try:
            merger.append(pdf)
        except:
            print('Could not append {}.'.format(pdf))
            e_count = e_count + 1

    with open(output_file, 'wb') as out:
        merger.write(out)

    if e_count == total_files:
        print('ERROR: No pdf was merged')
        return 0

    #merger.write(output_file)
    
    merger.close()

    return 1

def plot_image(path_to_image, ext = 0, output = None):
    """
    Plots and saves an image corresponding to a FITS image.

    Args:
        path_to_image: the path to the FITS file.
        ext: the extension of the image in the FITS file. 0 by default.
        output: None by default. Whethere or not save the image.

    Output:
        0 when finished.
    """

    with fits.open(path_to_image) as img:
        img_data = img[ext].data

    plt.imshow(img_data)

    if output is None:
        if os.getenv('DISPLAU'):
            plt.show()
        else:
            warnings.warn('Nothing was found in DISPLAY -- nothing was produced')
        return 0
    else:
        plt.savefig(output)

    return 0

def check_format_compatibility(out_format):
    """
    Checks if the input format is available as a matplotlib backend format.

    Args:
        out_format: the format to be checkec.

    Output:
        bool: whether or not the format can be used.
    """

    format_list = ('PNG', 'PDF', 'PGF', 'EPS', 'SVG2', 'RGBA', 'RAW', 'PS', 'SVG')

    if out_format.upper() in format_list:
        return True
    else:
        return False

def plot_region_box(ax, x1, y1, width, height, angle = 0, fill = False, colour = 'green', text = None, transform = None):
    '''
    Returns an axis object with with a box region printed on it.

    Args:
        ax: the original axis to use for the plot.
        x1: the starting x coordinate for the box.
        y1: the starting y coordinate for the box.
        width: the width of the box.
        height: the heigth of the box.
        angle: the angle of the box (if any). 0 by default.
        fill: boolean. Whether or not fill the box.
        colour: the colour of the box and the fill (if fill = True).
        text: None by default. Text to add at the (x1, y1) coordinates.
        transform: None by default. The transformation for the axis. Must be comatible with matplotlib standards.

    Output:
        f_ax: the axis after plotting the box.
    '''
     
    rec = Rectangle((x1, y1), width, height, angle, fill = fill, transform = transform, color = colour)
    
    pc = PatchCollection([rec], match_original = True)
    

    if text is None:
        pass
    else:
        ax.annotate(text, (x1, y1), color = colour, fontsize = 10)
        
    ax.add_collection(pc)
    
    return None

def reg_to_list(reg):
    ''' 
    Gets the information from a region based on the input string (based on ds9 formatting).

    So far, only tested for BOX regions.

    Args:
        reg: the region string in ds9 standards.

    Output:
        TODO: must be useful for all shapes. 
    ''' 
    
    shape_info, shape_details = reg.split('#') 
    shape_info = shape_info.strip().rstrip() 
    
    if 'BOX' in shape_info.upper(): 
        polygon = 'BOX' 
        x1, y1, width, heigth, angle = re.findall(r"[-+]?(?:\d*\.\d+|\d+)", shape_info)
        shape_info = (polygon, x1, y1, width, heigth, angle)
        
    # text details 
    if shape_details.find('COLOR') == -1: 
        colour = 'green' 
    else: 
        c_letters = shape_details[shape_details.find('COLOR') + 6 : shape_details.find('COLOR') + 9]
        if c_letters.upper() in 'RED': 
            colour = 'red' 
        elif c_letters.upper() in 'GREEN': 
            colour = 'green' 
        elif c_letters.upper() in 'BLUE': 
            ccolour = 'blue' 
        elif c_letters.upper() in 'CYAN': 
            colour = 'cyan' 
        elif c_letters.upper() in 'WHITE':  
            ccolour = 'white' 
        elif c_letters.upper() in 'MAGENTA': 
            colour = 'magenta' 
        elif c_letters.upper() in 'YELLOW': 
            ccolour = 'yellow' 
        else: 
            warnings.warn('Colour not recognised. Using green by default') 
            colour = 'green' 

    if not 'TEXT' in shape_details.upper(): 
        text = None
    else: 
        text = shape_details[shape_details.find('TEXT') + 6: shape_details.find('}')]
            
            
    return (polygon, float(x1), float(y1), float(width), float(heigth), float(angle), colour, text)



# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# eslewchainUtils

# Utilities for the eslewchain script.

import os
import subprocess
from astropy.io import fits

def delete_exposure_xtns(fill_file):
    """
    Removes exposure extensions from an event fle.
    
    Args:
        fill_file: name of the file.
    """

    with fits.open(fill_file, 'update') as fits_file:
        print('Deleting exposures from file: {}'.format(fill_file))
        try:
            del(fits_file['EXPOSU01'])
        except KeyError:
            print('Exposure 01 not deleted - not found')
        try:
            del(fits_file['EXPOSU02'])
        except KeyError:
            print('Exposure 02 not deleted - not found')
        try:
            del(fits_file['EXPOSU03'])
        except KeyError:
            print('Exposure 03 not deleted - not found')
        try:
            del(fits_file['EXPOSU04'])
        except KeyError:
            print('Exposure 04 not deleted - not found')
        try:
            del(fits_file['EXPOSU05'])
        except KeyError:
            print('Exposure 05 not deleted - not found')
        try:
            del(fits_file['EXPOSU06'])
        except KeyError:
            print('Exposure 06 not deleted - not found')
        try:
            del(fits_file['EXPOSU07'])
        except KeyError:
            print('Exposure 07 not deleted - not found')
        try:
            del(fits_file['EXPOSU08'])
        except KeyError:
            print('Exposure 08 not deleted - not found')
        try:
            del(fits_file['EXPOSU09'])
        except KeyError:
            print('Exposure 09 not deleted - not found')
        try:
            del(fits_file['EXPOSU10'])
        except KeyError:
            print('Exposure 10 not deleted - not found')
        try:
            del(fits_file['EXPOSU11'])
        except KeyError:
            print('Exposure 11 not deleted - not found')
        try:
            del(fits_file['EXPOSU12'])
        except KeyError:
            print('Exposure 12 not deleted - not found')
        fits_file.flush()

    print('Exposures deleted.')


def check_exp_extrns(event_file):
    """
    checks if all exposure extensions are present in the evt file
    Args:
        event_file: the event file.
    Output:
        str: an error if there are missing exposures.
    """

    exposure_count = 0
    with fits.open(event_file) as ev_file:
        for i in range(0, len(ev_file)):
            if 'EXPOSU' in ev_file[i].name:
                exposure_count = exposure_count + 1

    if exposure_count != 12:
        return 'The event file is missing at least one expousre.'
    else:
        print('Correct number of exposures.')


def rename_and_clean(obs, event_file):
    """
    Changes the name of some output files and removes other temporary files.
    
    Args:
        obs: observation number
        event_file: an event file.
    """

    att_file = 'P' + obs + 'OBX000ATTTSR0000.ds'
    os.system('mv ' + 'atthk.dat ' + '{}'.format(att_file))
    out_event_file = "P" + obs + "PNS" + "003" + "SLEVLI0000.ds"
    os.system('cp ' + event_file + ' ' + out_event_file)
    # once they have been renamed, now, remove the temporary files.
    os.system('rm ' + 'histo.fits ' + 'image.fits ' + 'spectrum.fits')




#!/usr/bin/python

from pysas.error import Error
from astropy.io import fits
import numpy as np
import sys
import os
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt


def getData(srcspecfilename,bkgspecfilename,expmapfilename,resmatfilename,intermediateFileFlag,intermediateFile):
    
    expmapfilename  = expmapfilename
    resmatfilename  = resmatfilename
    srcspecfilename = srcspecfilename
    bkgspecfilename = bkgspecfilename

    # Exposure map data and coordinate system values
    expmaphdul = fits.open(expmapfilename)
    
    naxis = expmaphdul[0].header['NAXIS1'] 
    crpix = expmaphdul[0].header['CRPIX1'] 
    crval = expmaphdul[0].header['CRVAL1']
    cdelt = expmaphdul[0].header['CDELT1']
    expmap = expmaphdul[0].data
    expmaphdul.close()
    
    #Read Response Matrix
    
    resmathdul = fits.open(resmatfilename)
    ebounds =resmathdul[2].data
    elow = ebounds['E_MIN']
    ehigh = ebounds['E_MAX']
    resmathdul.close()

    # Energy values
    enermat = 0.5 * (elow+ehigh)
    # Lambda grid
    minlow = crval-(cdelt/2.)
    maxlow = crval-(cdelt/2.)+(naxis*cdelt)
    minhigh = crval+(cdelt/2.)
    maxhigh = crval+(cdelt/2.)+(naxis*cdelt)
    bins = naxis
    llow = np.linspace(minlow, maxlow ,naxis)
    lhigh = np.linspace(minhigh, maxhigh ,naxis)
    lam = (llow+lhigh)/2.
    
    # Maximum expoure time per lambda channel
    exparray = np.max(expmap,0)
    # Maximum exposure time value
    maxexp = np.max(expmap)
 
    # Indexes where the expoosure time is less than 100 seconds and not equal to zero
    badcolumns  =  np.where((exparray != 0.) & (exparray <= 100.) )
    firstChannel = 0
    lastChannel  = 0
    # Loop to avoid channels at the begining and end with no exposure
    for i in range(len(exparray)):
        if (exparray[i] != 0) and (exparray[i] > maxexp*0.5):
            firstChannel = i
            break

    for i in  range( np.size(exparray) - 1, -1, -1):
        if (exparray[i] != 0) and (exparray[i] > maxexp*0.5):
            lastChannel = i
            break

    Error(client=os.path.basename(__file__), msgLayer='AppMsg', msgLevel='SilentMsg', msg=f'First channel with exposure time not equal to zero {firstChannel}').message()
    Error(client=os.path.basename(__file__), msgLayer='AppMsg', msgLevel='SilentMsg', msg=f'Last channel with exposure time not equal to zero {lastChannel}').message()
    
    # Indexes with 0s exposure time (chip gaps and bad columns)
    badchannels =  np.where( exparray[firstChannel:lastChannel] == 0)
    # shift the channels to the original grid
    badchannels = badchannels[0]+firstChannel
    
    jumps = np.diff(badchannels)
    jumpsind = np.where(jumps != 1)

    ccdChannels = np.empty((9,2),dtype=int)
 
    ccdChannels = np.insert(ccdChannels, 0, [firstChannel,badchannels[0]],axis=0)
    ccdChannels = np.delete(ccdChannels, (1), axis=0)
    gap = 1
    instrument = 1
    for i in range(np.size(jumpsind)):
        if instrument == 1 and i == 1:
            ccdChannels = np.insert(ccdChannels, i+gap, [650,965],axis=0)
            ccdChannels = np.delete(ccdChannels, (i+gap+1), axis=0)
            gap = 2
        ccdChannels = np.insert(ccdChannels, i+gap, [badchannels[jumpsind[0][i]],badchannels[jumpsind[0][i]+1]],axis=0)
        ccdChannels = np.delete(ccdChannels, (i+gap+1), axis=0)


    ccdChannels = np.insert(ccdChannels, 8, [badchannels[np.size(badchannels)-1],lastChannel],axis=0)  
    ccdChannels = np.delete(ccdChannels, (9), axis=0)
  

    goodchannels = np.where(exparray > (maxexp*0.85))
    lowexpchannels = np.where((exparray <= (maxexp*0.99)) & (exparray >= (maxexp*0.85)) )
    #print("GOOD CHANNELS ",np.size(goodchannels),np.size(badchannels))
    #print("Low level channels ",lowexpchannels)
    
    enerhigh =  12.398424 / llow
    enerlow = 12.398424 / lhigh

    x = np.linspace(0., np.size(goodchannels), np.size(goodchannels))
    #plt.plot(x, goodchannels[0], 'sb')
    #plt.show()
    #exit(1)
    #plt.imshow(expmap)
    #plt.xlim(1010,1080)
    #plt.ylim(0,160)
    #plt.show()
    #plt.plot(llow[badcolumns], exparray[badcolumns], 'sb')
    #plt.show()

    # I do not know what Jelle does with this channel...
    #print(np.where(np.isclose(exparray , 6793.85)))
    
    
    #Open Src Spec file
    srcspechdul = fits.open(srcspecfilename)
    irgs = srcspechdul[1].header['INSTRUME']
    irgs = int(irgs[-1])
    io = srcspechdul[1].header['RFLORDER']
    src = srcspechdul[1].data
    srccounts = src['COUNTS']
    srcbkgscale = src['BACKSCAL']
    #Open bkg Spec file
    bkgspechdul = fits.open(bkgspecfilename)
    bkg = bkgspechdul[1].data
    bkgcounts = bkg['COUNTS']
    bkgbkgscale = bkg['BACKSCAL']
    #bkgcounts = (bkgcounts*srcbkgscale)/bkgbkgscale
    a = (bkgcounts*srcbkgscale)
    kkk = np.zeros(len(bkgcounts))
    np.true_divide(a,bkgbkgscale, out=kkk, where=bkgbkgscale!=0) #only divide nonzeros
    bkgcounts = kkk

    bkgspechdul.close()
    srcspechdul.close()

    newsrccounts = (maxexp*srccounts[lowexpchannels])/exparray[lowexpchannels]
    newbkgcounts = (maxexp*bkgcounts[lowexpchannels])/exparray[lowexpchannels]
    #print("SRC BKG COUNTS ", srccounts[lowexpchannels],bkgcounts[lowexpchannels],exparray[lowexpchannels],maxexp,newbkgcounts)
    #plt.plot(elow, exparray, 'r')
    #plt.show()
    #plt.plot(elow[goodchannels], exparray[goodchannels], 'sr')
    #plt.plot(elow[lowexpchannels], exparray[lowexpchannels], 'sb')
    #plt.show()

    # Does not work... i do not know
    # pretend low exp channels are channels with maxexp
    bkgcounts[lowexpchannels] = newbkgcounts
    exparray[lowexpchannels]  = maxexp
    for i in range(np.size(lowexpchannels)):
        val1 = exparray[lowexpchannels[0][i]-1]
        val2 = exparray[lowexpchannels[0][i]+1]
        val = max(val1,val2)
        ind = lowexpchannels[0][i]
        exparray[ind] =  val
    #exparray[lowexpchannels] = maxexp
    #plt.plot(elow[goodchannels], exparray[goodchannels], 'sg')
    #plt.show()
    #plt.plot(elow, srccounts, 'sr')
    #plt.show()
    #plt.plot(elow[goodchannels], srccounts[goodchannels], 'sb')
    #plt.show()
    #plt.plot(elow, bkgcounts, 'sr')
    #plt.show()
    #plt.plot(elow[goodchannels], bkgcounts[goodchannels], 'sb')
    #plt.show()
    #print(elow[goodchannels],exparray[goodchannels],srccounts[goodchannels],bkgcounts[goodchannels])


    if (intermediateFileFlag == True):
        c1 = elow[goodchannels]
        c2 = ehigh[goodchannels]
        c3 = exparray[goodchannels]
        c4 = srccounts[goodchannels]/exparray[goodchannels]
        c5 = bkgcounts[goodchannels]/exparray[goodchannels]
        c4 = srccounts[goodchannels]
        c5 = bkgcounts[goodchannels]    
        c1 =  np.flipud(c1)
        c2 =  np.flipud(c2)
        c3 =  np.flipud(c3)
        c4 =  np.flipud(c4)
        c5 =  np.flipud(c5)
        col1 = fits.Column(name='ELow', format='D', array=c1)
        col2 = fits.Column(name='EHigh', format='D', array=c2)
        col3 = fits.Column(name='Exp_Time', format='D', array=c3)
        col4 = fits.Column(name='Source_Rate', format='D', array=c4)
        col5 = fits.Column(name='Back_Rate', format='D', array=c5)
    
        t = fits.BinTableHDU.from_columns([col1, col2, col3,col4,col5])
        t.writeto(intermediateFile,overwrite=True)

    
    return (irgs,io,elow[goodchannels],ehigh[goodchannels],exparray[goodchannels],srccounts[goodchannels],bkgcounts[goodchannels],goodchannels,lam[ccdChannels],ccdChannels)






#!/usr/bin/env python

#	ESA (C) 2000-2018 
#	This file is part of ESA's XMM-Newton Scientific Analysis System
#	(SAS).
#
#	SAS is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	SAS is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	   
#	You should have received a copy of the GNU General Public License
#	along with SAS.  If not, see <http://www.gnu.org/licenses/>.


from enum import Enum

class Instruments(Enum):
    EPN = 1
    EMOS1 = 2
    EMOS2 = 3
    RGS1 = 4
    RGS2 = 5 
    OM = 6
